from typing import List

from .device import Device, check_initialized

# Importing Dummy devices only for emulation purposes
import numpy as np
import pandas as pd
from datetime import datetime
from .dummy_heater import DummyHeater
from .dummy_motor import DummyMotor


# Given a parameter space of x1 and x2, bounded by x1_range and x2_range
# emulate a noisy potential energy surface as an objective function by 
# multiplying two fourier series from coefficients a1, b1 and a2, b2
# Then perform a "measurement" by sampling a point of the objective surface

# The two parameters x1, x2 in this case are a DummyHeater's temperature and a DummyMotor's speed
# This can be thought of as emulating a printing system where a material is processed at a certain temperature and speed
# Then the DummyMeter takes a "measurement" of the material and stores the data
class DummyMeter(Device):
    save_directory = 'data/dummy_meter/'

    def __init__(
            self, 
            name: str, 
            heater: DummyHeater,
            motor: DummyMotor,
            a1: List[float], 
            b1: List[float], 
            a2: List[float], 
            b2: List[float], 
            noise_width: float = 0.0
            ):
        super().__init__(name)
        # All arguments except 'name' are only for emulation purposes
        self.heater = heater
        self.motor = motor
        self.x1_range = (heater.min_temperature, heater.max_temperature)
        self.x2_range = (motor.motor.min_speed, motor.motor.max_speed)
        self.a1 = a1
        self.b1 = b1
        self.a2 = a2
        self.b2 = b2
        self.noise_width = noise_width

    def __init__(
            self, 
            name: str, 
            heater: dict,
            motor: dict,
            a1: List[float], 
            b1: List[float], 
            a2: List[float], 
            b2: List[float], 
            noise_width: float = 0.0
            ):
        super().__init__(name)
        # All arguments except 'name' are only for emulation purposes
        self.heater = DummyHeater(**heater)
        self.motor = DummyMotor(**motor)
        self.x1_range = (self.heater.min_temperature, self.heater.max_temperature)
        self.x2_range = (self.motor.motor.min_speed, self.motor.motor.max_speed)
        self.a1 = a1
        self.b1 = b1
        self.a2 = a2
        self.b2 = b2
        self.noise_width = noise_width

    def get_args(self) -> dict:
        args_dict = {
            "name": self._name,
            "heater": self.heater.get_args(),
            "motor": self.motor.get_args(),
            "a1": self.a1,
            "b1": self.b1,
            "a2": self.a2,
            "b2": self.b2,
            "noise_width": self.noise_width,
        }
        return args_dict

    def initialize(self):
        self._is_initialized = True
        return (True, "Initialized DummyMeter")

    def deinitialize(self, reset_init_flag: bool = True):
        if reset_init_flag:
            self._is_initialized = False
        return (True, "Deinitialized DummyMeter")

    @check_initialized
    def measure(self, filename: str = None):
        # if not self._is_initialized:
        #     return (False, "DummyMeter is not initialized")

        # Emulating dependency on temperature and speed
        x1 = self.heater.temperature
        x2 = self.motor.speed

        # Take a "measurement"
        data = self.fourier2d(
            [x1, x2],
            self.x1_range,
            self.x2_range,
            self.a1,
            self.b1,
            self.a2,
            self.b2,
            self.noise_width
            )
        
        # Store the measured data to a csv file
        if filename is None:
            timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            filename = timestamp

        dataframe = pd.DataFrame()
        dataframe['Temperature'] = [x1]
        dataframe['Speed'] = [x2]
        dataframe['Data'] = [data]
        
        fullfilename = self.save_directory + filename + '.csv'
        dataframe.to_csv(fullfilename, mode='w', index=False)

        return (True, "Measured data and saved to file at " + fullfilename)
    
    @staticmethod
    def fourier2d(x, x1_range, x2_range, a1, b1, a2, b2, noise_width) -> float:
        x1 = x[0]
        x2 = x[1]
        x1_min = x1_range[0]
        x1_max = x1_range[1]
        x2_min = x2_range[0]
        x2_max = x2_range[1]
        w1 = 2 * np.pi / (x1_max - x1_min)
        w2 = 2 * np.pi / (x2_max - x2_min)

        f1 = 0
        f2 = 0

        for i in range(a1.shape[0]):
            f1 += a1[i]*np.cos(i*x1*w1) + b1[i]*np.sin(i*x1*w1)
            f2 += a2[i]*np.cos(i*x2*w1) + b2[i]*np.sin(i*x2*w1)

        f = f1 * f2
        if noise_width:
            f += np.random.normal(0, noise_width)
        return f[0]




from typing import Optional, Tuple, Union
import serial
import functools

from .device import ArduinoSerialDevice, check_initialized, check_serial


def check_stepper_num(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        # could be a kwarg or arg
        if 'stepper_number' in kwargs:
            stepper_number = kwargs['stepper_number']
        else:
            stepper_number = args[0]

        if not self.is_stepper_num_valid(stepper_number):
            return (False, "Stepper number is not valid or not part of passed tuple during construction.")
        return func(self, *args, **kwargs)
    return wrapper


class MultiStepper(ArduinoSerialDevice):
    def __init__(
            self,
            name: str, 
            port: str, 
            baudrate: int, 
            timeout: Optional[float] = 1.0, 
            stepper_list: Tuple[int, ...] = (1,),
            move_timeout: float = 30.0):
        
        super().__init__(name, port, baudrate, timeout)
        self._stepper_list = stepper_list
        self._move_timeout = move_timeout # max move timeout for ALL possible motors controlled by arduino

    def initialize(self):
        for stepper in self._stepper_list:
            was_homed, comment = self.home(stepper)
            if not was_homed:
                self._is_initialized = False
                return (was_homed, comment)

        self._is_initialized = True
        return (True, "All stepper motors " + str(self._stepper_list) + " successfully homed and initialized. All current positions set to zero.")

    def deinitialize(self, reset_init_flag: bool = True):
        for stepper in self._stepper_list:
            was_zeroed, comment = self.move_absolute(stepper, 0.0)
            if not was_zeroed:
                return (was_zeroed, comment)

        if reset_init_flag:
            self._is_initialized = False

        return (True, "All stepper motors " + str(self._stepper_list) + " successfully deinitialized by moving to position zero.")
    
    @check_serial
    @check_stepper_num
    def home(self, stepper_number: int) -> Tuple[bool, str]:
        # if not self.ser.is_open:    
        #     return (False, "Serial port " + self._port + " is not open. ")
        # if not self.is_stepper_num_valid(stepper_number):
        #     return (False, "Stepper number is not valid or not part of passed tuple during construction.")

        command = ">hm " + str(stepper_number) + "\n"
        self.ser.write(command.encode('ascii'))
        return self.check_ack_succ(ack_timeout=1, succ_timeout=self._move_timeout)

    @check_serial
    @check_initialized
    @check_stepper_num
    def move_absolute(self, stepper_number: int, position: float) -> Tuple[bool, str]:
        # if not self.ser.is_open:    
        #     return (False, "Serial port " + self._port + " is not open. ")
        # elif not self.is_stepper_num_valid(stepper_number):
        #     return (False, "Stepper number is not valid or not part of passed tuple during construction.")
        # elif not self._is_initialized:
        #     return (False, "Stepper motor " + str(stepper_number) + " is not initialized.")
        # else:
        # if not self.is_stepper_num_valid(stepper_number):
        #     return (False, "Stepper number is not valid or not part of passed tuple during construction.")

        command = ">mv " + str(stepper_number) + " " + str(position) + "\n"
        self.ser.write(command.encode('ascii'))
        return self.check_ack_succ(ack_timeout=1, succ_timeout=self._move_timeout)

    @check_serial
    @check_initialized
    @check_stepper_num
    def move_relative(self, stepper_number: int, distance: float) -> Tuple[bool, str]:
        # if not self.ser.is_open:    
        #     return (False, "Serial port " + self._port + " is not open. ")
        # elif not self.is_stepper_num_valid(stepper_number):
        #     return (False, "Stepper number is not valid or not part of passed tuple during construction.")
        # elif not self._is_initialized:
        #     return (False, "Stepper motor " + str(stepper_number) + " is not initialized.")
        # else:
        # if not self.is_stepper_num_valid(stepper_number):
        #     return (False, "Stepper number is not valid or not part of passed tuple during construction.")

        command = ">mvr " + str(stepper_number) + " " + str(distance) + "\n"
        self.ser.write(command.encode('ascii'))
        return self.check_ack_succ(ack_timeout=1.0, succ_timeout=self._move_timeout)

    @check_serial
    @check_initialized
    @check_stepper_num
    def position(self, stepper_number: int) -> Tuple[bool, Union[float, str]]:
        # if not self.ser.is_open:    
        #     return (False, "Serial port " + self._port + " is not open. ")
        # elif not self.is_stepper_num_valid(stepper_number):
        #     return (False, "Stepper number is not valid or not part of passed tuple during construction.")
        # elif not self._is_initialized:
        #     return (False, "Stepper motor " + str(stepper_number) + " is not initialized.")
        # else:
        # if not self.is_stepper_num_valid(stepper_number):
        #     return (False, "Stepper number is not valid or not part of passed tuple during construction.")

        command = ">wm " + str(stepper_number) + "\n"
        self.ser.write(command.encode('ascii'))
        was_successful, comment = self.check_ack_succ()

        if not was_successful:
            return (was_successful, comment)

        has_position, position_str = super().parse_equal_sign(comment)

        if not has_position:
            return (has_position, "Message from device did not contain motor position.")
        
        return (True, float(position_str))
 
    def is_stepper_num_valid(self, stepper_number: int) -> bool:
        if stepper_number in self._stepper_list:
            return True
        else:
            return False


# The arduino program is able to check if the passed stepper number is valid or not and return True/False
# Therefore, I initially concluded that there is no need to validate the stepper number argument in the methods here or pass it during construction
# However, suppose there are (1, 2, 3) steppers and we only pass (1,2) to initialize_all, then the MultiStepper object will be flagged as 
# being initialized and it will be possible to send commands to stepper #3. Although the arduino may be able to prevent moving stepper #3 if its not homed
# it is technically possible for the command to be valid from the host's perspective which in this particular case should not happen even if the arduino will catch it
# Also suppose the arduino doesn't have a check at it's level then we will be able to move stepper #3 without properly initializing it
# therefore we should declare during construction which steppers we plan to use with an immutable tuple 
# and still check that a passed stepper_number argument is always in that tuple

# currently the way we determine if the movement completed or not is by waiting for the arduino to send a SUCC confirmation once it reaches its target
# this means that the host side needs to implement a timeout longer than the movement duration, either using a blanket large value or based on the speed, distance, and a safe extra time
# At the moment a blanket timeout value is used based on the known stepper application
# Other better or worse ways would be the following:
# -poll the position of the arduino repeatedly until it matches with the desired target position. 
#   However, if the arduino gets stuck this could infinite loop meaning we need another timeout
#   Additionally if there is small error between target and acceptable position, that needs to be accounted for and just adds more complexity
# -poll whether the arduino is moving or not. Better than polling position but what if arduino never moves or never stops for some reason. 
#   this means we may still need to implement a timeout and also have the arduino spit out error messages for these cases
#
# Regarding the speed of the stepper motor, at the moment it is coded into the Arduino. It can easily be changed to a modifiable parameter set by a serial command
# And then we can control it from here. But at the moment this is not implemented as my current use case for custom stepper motors is purely for positioning purposes only


"""Requires download of NI-VISA with equivalent bitness"""
import pyvisa
from time import sleep
from datetime import datetime
import pandas as pd
from typing import Optional, Tuple
from .device import Device, check_initialized



class Keithley2450(Device):
    save_directory = 'data/keithley_2450/'

    def __init__(self, name: str, ID: str, query_delay: int):
        super().__init__(name)
        self.ID = ID
        self.query_delay = query_delay
        self.keithley = pyvisa.ResourceManager().open_resource(self.ID)
        print(self.keithley.query('*IDN?'))
        sleep(self.query_delay)

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self.name,
            "ID": self.ID,
            "query_delay": self.query_delay,
        }
        return args_dict
    
    def update_init_args(self, args_dict: dict):
        self.name = args_dict["name"]
        self.ID = args_dict["ID"]
        self.query_delay = args_dict["query_delay"]

    @property   #TODO: Check if necessary
    def terminal_pos(self) -> str:
        return self.position
    
    def initialize(self) -> Tuple[bool, str]:
        self.keithley.write('*RST')
        self._is_initialized = True
        return (True, "Initialized Keithley2450 by performing device reset")

    def deinitialize(self) -> Tuple[bool, str]:
        #TODO: Keithley deinitialization steps
        self.keithley.close()
        self._is_initialized = False
        return(True, "Deinitialized Keithley2450")
    
    @check_initialized
    def wait(self):
        self.keithley.query('*OPC?')
        if self.keithley.query('*ESR?') != 1: 
            self.keithley.write('*CLS')
            print("Wait Completed, Standard Event Status Register cleared")

    #TODO: Test wait function below
    """
    def wait(self):
        self.keithley.write('*OPC?')
        while True:
            try: 
                self.keithley.read()
                break
            except pyvisa.errors.VisaIOError:
                continue
            """

    @check_initialized
    def write_command(self, command: str):
        self.keithley.write(command)
        
    @check_initialized
    def terminal_pos(self, position: str = 'front'):
        """Set terminal positions to front or rear"""
        if position == "rear":
            self.keithley.write('ROUT:TERM REAR')
            self.wait()
        elif position == "front":
            self.keithley.write('ROUT:TERM FRONT')
            self.wait()
        else:
            raise Exception(f"Expected 'rear' or 'front', found {position}")
    @check_initialized
    def error_check(self):
        """Check for errors raised by Keithley2450"""
        self.keithley.write('SYSTem:ERRor:COUNt?')
        num_errors = int(self.keithley.read())
        if num_errors != 0:
            errors = []
            for i in range(num_errors):
                self.keithley.write('SYSTem:ERRor:NEXT?')
                errors.append(self.device.read())
            errors = ''.join(errors)
            raise Warning(f"An error has occurred:\n {errors}")
    
    @check_initialized
    def clear_buffer(self, buffer: str = 'defbuffer1') -> Tuple[bool, str]:
        """Clear the data storage buffer within the Keithley2450"""
        self.keithley.write(f':TRACe:CLEar "{buffer}"')
        return (True, f'Cleared buffer {buffer}')

        
    @check_initialized
    def IV_characteristic(self, ilimit: float, vmin: float, vmax: float, delay: float, steps: int = 60):
        """Sourcing voltage and measuring current with linear sweep"""
        if ilimit <= 1e-9 and ilimit >= 1.05:
            raise Exception(f"Expected source current limit between 1nA and 1.05A")

        if vmin < -210 or vmin > 210:
            raise Exception(f"Voltage minimum out of range -210V to 210V")
        
        if vmax < -210 or vmin > 210:
            raise Exception(f"Voltage maximum out of range -210V to 210V")
        if vmax < vmin:
            raise Exception(f"Voltage minimum is greater than voltage maximum")

        if delay < 50e-6:
            raise Exception(f"Delay value too small, must be greater than 50 µs")
        
        self.keithley.write('*RST')
        self.keithley.write('SENS:CURR:RANG:AUTO ON')
        self.keithley.write('SYST:RSEN ON')
        self.keithley.write(f"SOURce:VOLT:ILIMit {ilimit}")
        self.keithley.write(f"SOURce:SWE:VOLT:LIN {vmin}, {vmax}, {steps}, {delay}")
        self.keithley.write('INIT')
        self.keithley.write('*WAI')

    @check_initialized
    def four_point(self, test_curr: float, vlimit: float, curr_reversal: bool = False) -> float:
        """Measure resistance through four-point collinear probe method"""

        self.keithley.write('*RST')
        self.keithley.write('SENS:VOLT:RANG:AUTO ON')
        self.keithley.write('SYST:RSEN ON')
        self.keithley.write(f"SOURce:CURR:VLIMit {vlimit}")
        self.keithley.write(f"SOURce:CURR {test_curr}")
        self.keithley.write(':OUTP ON')
        self.keithley.write('*WAI')
        sleep(2)
        res1 = self.keithley.query(':READ?')
        self.keithley.write(':OUTP OFF')
        four_point_data = res1
        if curr_reversal:
            self.keithley.write(f"SOURce:CURR:VLIMit {vlimit}")
            self.keithley.write(f"SOURce:CURR {-1*test_curr}")
            self.keithley.write(':OUTP ON')
            self.keithley.write('*WAI')
            sleep(2)
            res2 = self.keithley.query('READ?')
            self.keithley.write(':OUTP OFF')
            four_point_data = (res1 + res2) / 2
        return four_point_data
    
        
    @check_initialized
    def get_data(self, filename: str = None, four_point: bool = False):
        """Retrieve data from buffer and save locally"""

        if not four_point:
            self.keithley.write(':TRACe:ACTual:END?')
            end_index = int(self.keithley.read())
            self.keithley.write(f':TRACe:DATA? 1, {end_index}, "defbuffer1", RELative, SOURce, READing')
            results = self.keithley.read()
            results = results.split(',')
            results = list(map(float, results))
            data = {
                'time' : results[::3],
                'source' : results[1::3],
                'reading' : results[2::3]
            }
        else:
            self.keithley.write(':TRACe:ACTual:END?')
            end_index = int(self.keithley.read())
            self.keithley.write(f':TRACe:DATA? 1, {end_index}, "defbuffer1"')
            data = self.keithley.read()

        if filename is None:
            timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            filename = timestamp

        dataframe = pd.DataFrame(data)

        fullfilename = self.save_directory + filename + '.csv'
        dataframe.to_csv(fullfilename, mode = 'w', index = False)

        return (True, "Saved data to at " + fullfilename)
    




from typing import Union, Tuple, Dict, List, Optional
from datetime import datetime

import numpy as np
import pandas as pd
from scipy.interpolate import interp1d
from scipy.optimize import minimize
import stellarnet_driver3 as sn

from .device import Device, check_initialized

# TODO
# convert to serial device parent when arduino added
# add shutter and lamp control
# check slicing as second index is not included in slice
# type hinting of ndarrays?
# type hint the static methods?

class StellarNetSpectrometer(Device):
    save_directory = 'data/spectroscopy/'

    def __init__(self, name: str, spec_keys: List[str] = ['UV-Vis', 'NIR']):
        super().__init__(name)
        self.spectrometer_dict = {}
        self.wavelength_dict = {}
        self.dark_spectra_dict = {}
        self.blank_spectra_dict = {}
        self.absorbance_dict = {}
        self.merged_absorbance = None
        self.num_spectrometers = 0
        self.spec_keys = spec_keys

    @staticmethod
    def num_specs_connected() -> int:
        num_connected = 0
        start_wav = [-1.]
        while True:
            try:
                spec, wav = sn.array_get_spec(num_connected)
                start_wav.append(wav[0].item())
                if start_wav[num_connected+1] == start_wav[num_connected]:
                    break
                else:
                    num_connected += 1
            except:
                break
        return num_connected

    def initialize(self) -> Tuple[bool, str]:
        #lamp on
        #shutter in/out
        #any other Arduino initialization

        self.num_spectrometers = self.num_specs_connected()
        if self.num_spectrometers == 0:
            self._is_initialized = False
            return (False, "There are no spectrometers connected")

        for ndx in range(self.num_spectrometers):
            spectrometer, wavelengths = sn.array_get_spec(ndx)
            if wavelengths[0] < 200.0:
                # UV-VIS
                self.spectrometer_dict['UV-Vis'] = spectrometer
                self.wavelength_dict['UV-Vis'] = wavelengths
            else:
                # NIR
                self.spectrometer_dict['NIR'] = spectrometer
                self.wavelength_dict['NIR'] = wavelengths
            # For additional spectrometers add to the if else ladder

        # Check that we were able to initialize all spectrometers that were declared during construction
        if set(self.spec_keys) == set(self.spectrometer_dict.keys()):        
            self._is_initialized = True
            return (True, "Successfully initialized " + str(self.num_spectrometers) + " spectrometers: " + str(list(self.spectrometer_dict.keys())))
        else:
            self._is_initialized = False
            return (False, "Not all declared spectrometers were initialized. Only initialized: " + str(list(self.spectrometer_dict.keys())))

    def deinitialize(self, reset_init_flag: bool = True) -> Tuple[bool, str]:
        # turn off lamp?
        # move shutter the position  so that initialize can  be ready to move the shutter  correct position when starting the instrument
        if reset_init_flag:
            self._is_initialized = False

        return (True, "Pass, nothing to deinitialize for now.")

    @check_initialized
    def get_spectrum_counts(
            self, 
            spec_key: str, 
            integration_time: int = 100, 
            scans_to_avg: int = 3,  
            smoothing: int = 0, 
            xtiming: int = 1) -> Tuple[bool, str]:

        # if not self._is_initialized:
        #     return (False, "Spectrometer system not initialized")

        if spec_key in self.spectrometer_dict.keys():
            self.spectrometer_dict[spec_key]['device'].set_config(
                int_time=integration_time, 
                scans_to_avg=scans_to_avg, 
                x_smooth=smoothing, 
                x_timing=xtiming)

            spectrum_array = sn.array_spectrum(self.spectrometer_dict[spec_key], self.wavelength_dict[spec_key])

            max_count = np.amax(spectrum_array[:,1], axis=0)
            print("Max count: " + str(max_count))
            if max_count > 65500:
                return (False, spec_key + " detector is saturated, lower integration time")
            return (True, spectrum_array)
        else:
            return (False, spec_key + " spectrometer is not found" )

    def get_all_spectra_counts(
            self, 
            integration_times: Tuple[int, ...] = (100, 100), 
            scans_to_avg: Tuple[int, ...] = (3, 3), 
            smoothings: Tuple[int, ...] = (0, 0), 
            xtimings: Tuple[int, ...] = (1, 1)) -> Tuple[bool, str]:

        # Modify the variable 'spec_keys' if you don't intend to use all spectrometers
        if self.num_spectrometers != len(self.spec_keys):
            return (False, "Spectrometers are not all connected")

        spectrum_array_dict = {}
        # print(spec_keys)
        # using self.spec_keys to ensure that the order within the parameter tuple matches the order of the declared spec_keys
        for ndx, spec_key in enumerate(self.spec_keys):
            # this function should already check if the spec_key is valid
            result, spectrum_array = self.get_spectrum_counts(
                                        spec_key,
                                        integration_time=integration_times[ndx], 
                                        scans_to_avg=scans_to_avg[ndx], 
                                        smoothing=smoothings[ndx], 
                                        xtiming=xtimings[ndx])
            if not result:
                return result, spectrum_array

            spectrum_array_dict[spec_key] = spectrum_array

        return (True, spectrum_array_dict)


    def update_all_dark_spectra(
            self, 
            integration_times: Tuple[int, ...] = (100, 100), 
            scans_to_avg: Tuple[int, ...] = (3, 3), 
            smoothings: Tuple[int, ...] = (0, 0), 
            xtimings: Tuple[int, ...] = (1, 1)) -> Tuple[bool, str]:

        result, spectrum_array_dict = self.get_all_spectra_counts(
                                        integration_times, 
                                        scans_to_avg, 
                                        smoothings, 
                                        xtimings)

        if not result:
            return result, spectrum_array_dict

        for key, value in spectrum_array_dict.items():
            self.dark_spectra_dict[key] = value

        return (True, "All dark spectra stored")

    def update_all_blank_spectra(
            self, 
            integration_times: Tuple[int, ...] = (100, 100), 
            scans_to_avg: Tuple[int, ...] = (3, 3), 
            smoothings: Tuple[int, ...] = (0, 0), 
            xtimings: Tuple[int, ...] = (1, 1)) -> Tuple[bool, str]:

        result, spectrum_array_dict = self.get_all_spectra_counts(
                                        integration_times, 
                                        scans_to_avg, 
                                        smoothings, 
                                        xtimings)

        if not result:
            return result, spectrum_array_dict

        for key, value in spectrum_array_dict.items():
            self.blank_spectra_dict[key] = value

        return (True, "All blank spectra stored")

    def get_all_absorbance(
            self,
            save_to_file: bool = False, 
            filename: Optional[str] = None,
            integration_times: Tuple[int, ...] = (100, 100), 
            scans_to_avg: Tuple[int, ...] = (3, 3), 
            smoothings: Tuple[int, ...] = (0, 0), 
            xtimings: Tuple[int, ...] = (1, 1)) -> Tuple[bool, str]:

        # get the sample spectra
        result, spectrum_array_dict = self.get_all_spectra_counts(
                                        integration_times, 
                                        scans_to_avg, 
                                        smoothings, 
                                        xtimings)
        if save_to_file and filename is None:
            timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
            filename = timestamp


        # check if failed
        if not result:
            return result, spectrum_array_dict

        # check that dark/blank spectra exists for each initialized spectrometer
        for key in self.spectrometer_dict.keys():
            if not key in self.blank_spectra_dict:
                return (False, "Blank spectra for " + key + " is missing")
            if not key in self.dark_spectra_dict:
                return (False, "Dark spectra for " + key + " is missing")

        # Calculate absorbance for each spectra, save to self
        # Absorbance is -log10((Isample - Idark)/(Iblank - Idark))
        for ndx, spec_key in enumerate(self.spec_keys):
            wavelength = self.wavelength_dict[spec_key].copy()
            dark_spec = self.dark_spectra_dict[spec_key][:,1].copy()
            blank_spec = self.blank_spectra_dict[spec_key][:,1].copy()
            sam_spec = spectrum_array_dict[spec_key][:,1].copy()

            absorbance = np.expand_dims(-np.log10((sam_spec - dark_spec) / (blank_spec - dark_spec)), axis=1)
            absorbance_array = np.hstack((wavelength, absorbance))

            self.absorbance_dict[spec_key] = absorbance_array
            
            # save all data related to absorbance calculation to a file per spectrometer
            if save_to_file:
                data = pd.DataFrame()
                data[spec_key + ' Wavelength'] = np.squeeze(wavelength)
                data[spec_key + ' Dark Counts'] = np.squeeze(dark_spec)
                data[spec_key + ' Blank Counts'] = np.squeeze(blank_spec)
                data[spec_key + ' Sample Counts'] = np.squeeze(sam_spec)
                data[spec_key + ' Absorbance'] = np.squeeze(absorbance)
                
                comment = ["# spec_key = " + spec_key + "\n",
                            "# integration_time = " + str(integration_times[ndx]) + "\n",
                            "# scans_to_avg = " + str(scans_to_avg[ndx]) + "\n",
                            "# smoothing = " + str(smoothings[ndx]) + "\n",
                            "# xtiming = " + str(xtimings[ndx]) + "\n"]
                
                fullfilename = self.save_directory + filename + '_' + spec_key + '.csv'
                with open(fullfilename, 'w') as file:
                    file.writelines(comment)
                data.to_csv(fullfilename, mode='a', index_label='Index')

        if save_to_file:
            all_comments = ["# spec_keys = " + str(self.spec_keys) + "\n"
                            "# integration_times = " + str(integration_times) + "\n",
                            "# scans_to_avg = " + str(scans_to_avg) + "\n",
                            "# smoothings = " + str(smoothings) + "\n",
                            "# xtimings = " + str(xtimings) + "\n"]
        else:
            all_comments = ['#\n',]
        # Merge the absorbance spectra and optionally save to file
        # What happens if only 1 spectrometer is connected/being used?
        if len(self.spec_keys) > 1:
            result, message = self.merge_absorbance(save_to_file, filename, all_comments)

            if not result:
                return result, message

        if save_to_file:
            return (True, "All absorbance spectra stored to instance and saved to file: " + self.save_directory + filename)
        else:
            return (True, "All absorbance spectra stored to instance but not saved to file")

    # hard coded for UV-Vis and NIR
    # what happens if only 1 spectrometer is connected/being used?
    def merge_absorbance(self, save_to_file: bool, filename: str, comment_list: List[str]) -> Tuple[bool, str]:
        uv_wavelength = np.squeeze(self.wavelength_dict['UV-Vis'].copy())
        nir_wavelength = np.squeeze(self.wavelength_dict['NIR'].copy())
        uv_absorbance = np.squeeze(self.absorbance_dict['UV-Vis'][:,1].copy())
        nir_absorbance = np.squeeze(self.absorbance_dict['NIR'][:,1].copy())

        # start and end of overlapping regions
        WAVE_START = 900.0
        WAVE_END = 1030.0

        merge_result = minimize(self.merge_error, [1, 0], args=(uv_wavelength, uv_absorbance, nir_wavelength, nir_absorbance, WAVE_START, WAVE_END), method='BFGS')

        if not merge_result:
            return (False, "Failed to merge UV-Vis and NIR absorbance spectra: " + merge_result.message)
        else:
            scale = merge_result.x[0]
            shift = merge_result.x[1]

            # uv is not modified, nir is adjusted to match uv
            uv_array = self.absorbance_dict['UV-Vis'].copy()
            nir_wavelength = self.wavelength_dict['NIR'].copy()

            nir_absorbance = np.expand_dims(self.scale_shift_data([scale, shift], self.absorbance_dict['NIR'][:,1].copy()), axis=1)
            nir_array = np.hstack((nir_wavelength, nir_absorbance))

            UV_START = 210.0
            UV_END = 1030.0
            NIR_START = 900.0
            NIR_END = 1700.0

            uv_array = self.truncate_ends_by_wavelength(uv_array, UV_START, UV_END)
            nir_array = self.truncate_ends_by_wavelength(nir_array, NIR_START, NIR_END)

            merged_array = np.vstack((uv_array, nir_array))
            merged_array = merged_array[merged_array[:,0].argsort()]
            self.merged_absorbance = merged_array
            
            if save_to_file:
                data = pd.DataFrame()
                data['Wavelength'] = np.squeeze(self.merged_absorbance[:,0].copy())
                data['Absorbance'] = np.squeeze(self.merged_absorbance[:,1].copy())

                fullfilename = self.save_directory + filename + '_merged.csv'
                comment_list.append("# To merge, NIR data is scaled first then shifted\n")
                comment_list.append("# scale = " + str(scale) + "\n")
                comment_list.append("# shift = " + str(shift) + "\n")
                with open(fullfilename, 'w') as file:
                    file.writelines(comment_list)
                data.to_csv(fullfilename, mode='a', index_label='Index')

            return (True, "Successfully merged UV-Vis and NIR absorbance spectra: " + merge_result.message)

    # y and return are ndarrays
    @staticmethod
    def scale_shift_data(params: List[float], y):
        scale = params[0]
        shift = params[1]
        return y * scale + shift
        # return (y + shift) * scale

    @staticmethod
    def truncate_ends_by_wavelength(array, start: float, end: float):
        # expects array to be two columns with first being wavelength and second being the data of interest (counts, absorbance, etc.)
        return array[np.logical_and(array[:,0]>start, array[:,0]<end)]

    @staticmethod
    def find_nearest(x, x0):
        ndx = np.abs(x - x0).argmin()
        return ndx, x[ndx]

    @staticmethod
    def merge_error(params, x1, y1, x2, y2, x_start, x_end):
        # y1 (UV-Vis) has higher resolution (2048) than y2 (NIR, 512)
        # so although we will adjust y2 to match y1, y1 will be interpolated onto x2
        # get the overlapped slice of the data that is not being interpolated (x2)
        x2_start_ndx = StellarNetSpectrometer.find_nearest(x2, x_start)[0]
        x2_end_ndx = StellarNetSpectrometer.find_nearest(x2, x_end)[0]
        x2_slice = x2[x2_start_ndx : x2_end_ndx].copy()
        # adjust y2
        new_y2 = StellarNetSpectrometer.scale_shift_data(params, y2)
        y2_slice = new_y2[x2_start_ndx : x2_end_ndx].copy()
        # get interpolated function of y1(x1)
        y1_interp = interp1d(x1, y1)
        # get y1 interpolated onto x2
        y1_slice = y1_interp(x2_slice)
        # calc error
        sum_squares = ((y1_slice - y2_slice) ** 2).sum()
        return sum_squares

    def servo_shutter_in(self):
        pass

    def servo_shutter_out(self):
        pass

    def lamp_relay_on(self):
        pass

    def lamp_relay_off(self):
        pass


# Considered making get_spectrum_counts generalized to accept either a single spectrometer (spec_key + params) 
# or any sized list of spec_keys + list of params.
# I ended up not doing it this way and making get_spectrum_counts take only 1 spectrometer
# Then I made a second function get_all_spectra_counts that calls get_spectrum_counts multiple times.
# My subsequent functions to get dark spectra, blank spectra, absorbance spectra then work with get_all_spectra_counts
# At first this seems more complicated. However, I did it to avoid list/tuple type checking and len() checking for every single param
# and because I wanted to avoid the following situation:
# 1) User gets dark and blank spectra for all spectrometers and proceeds to use them to get absorption spectras
# 2) Later, user needs needs to retake the dark/blank spectra but makes an error and uses the generalized methods for NOT all spectrometers
# 3) User proceeds to calculate all absorption spectra for all spectrometers
# In this case, not all spectrometers had their dark/blank spectra correctly retaken.
# The data is then incorrect for those spectrometers since they use the old dark/blank but the user has no idea this just happened.
# Another option is to null all spectrometer dark/blanks when even only retaking for 1 spectrometer, but requires implementing checks and error messages
# It is easier at the moment to just enforce all spectrometers are used all the time by writing methods that always use all spectrometers
# However, The "get_all_..." methods can generalize to 1 spectrometer based on the spec_keys arg passed during construction


import sys
import logging

class ConsoleInterceptor:
    def __init__(self):
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        self.intercepted_messages = []

    def start_interception(self):
        sys.stdout = self.InterceptedStream(self.stdout, self._handle_message)
        sys.stderr = self.InterceptedStream(self.stderr, self._handle_message)
        self._configure_logger()

    def stop_interception(self):
        sys.stdout = self.stdout
        sys.stderr = self.stderr

    def get_intercepted_messages(self):
        return self.intercepted_messages

    def _handle_message(self, message):
        # Store the intercepted message for later use
        self.intercepted_messages.append(message)

    def _configure_logger(self):
        logger = logging.getLogger()
        logger.addHandler(self._get_logging_handler())
        logger.addFilter(self._get_logging_filter())

    def _get_logging_handler(self):
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter('%(levelname)s: %(message)s'))
        return handler

    def _get_logging_filter(self):
        return self.InterceptedMessageFilter(self.intercepted_messages)

    class InterceptedStream:
        def __init__(self, stream, handler):
            self.stream = stream
            self.handler = handler

        def write(self, message):
            self.handler(message)
            self.stream.write(message)

        def flush(self):
            self.stream.flush()

    class InterceptedMessageFilter(logging.Filter):
        def __init__(self, intercepted_messages):
            super().__init__()
            self.intercepted_messages = intercepted_messages

        def filter(self, record):
            message = record.getMessage()
            return message not in self.intercepted_messages




from .command import Command, CompositeCommand
from devices.dummy_heater import DummyHeater
from devices.dummy_motor import DummyMotor
from commands.dummy_heater_commands import *
from commands.dummy_motor_commands import *

# Example of a composite command that interacts with multiple devices (receivers)
class DummyPrinterHeatMove(CompositeCommand):
    """Sets DummyHeater to temperature, moves DummyMotor a relative distance and back, and sets DummyHeater to 25, reverts heat_rate and speed to original values"""
    def __init__(
            self,
            heater: DummyHeater, 
            motor: DummyMotor, 
            temp: float, 
            heat_rate: float, 
            distance: float, 
            speed: float, 
            **kwargs):
        super().__init__(**kwargs)
        initial_heat_rate = heater.heat_rate
        initial_speed = motor.speed
        self.add_command(DummyHeaterSetHeatRate(heater, heat_rate))
        self.add_command(DummyHeaterSetTemp(heater, temp))
        self.add_command(DummyMotorSetSpeed(motor, speed))
        self.add_command(DummyMotorMoveRelative(motor, distance))
        self.add_command(DummyMotorMoveRelative(motor, -distance))
        self.add_command(DummyHeaterSetTemp(heater, 25))
        self.add_command(DummyHeaterSetHeatRate(heater, initial_heat_rate))
        self.add_command(DummyMotorSetSpeed(motor, initial_speed))



from devices.device import Device
from .command import Command, CommandResult
from devices.linear_stage_150 import LinearStage150


class LinearStage150ParentCommand(Command):
    """Parent class for all LinearStage150 commands."""
    receiver_cls = LinearStage150

    def __init__(self, receiver: LinearStage150, **kwargs):
        super().__init__(receiver, **kwargs)

class LinearStage150Connect(LinearStage150ParentCommand):
    """Open a serial port for the linear stage."""

    def __init__(self, receiver: LinearStage150, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.start_serial()) 

class LinearStage150Initialize(LinearStage150ParentCommand):
    """Initialize the linear stage by homing it."""

    def __init__(self, receiver: LinearStage150, **kwargs):
        super().__init__(receiver, **kwargs)
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class LinearStage150Deinitialize(LinearStage150ParentCommand):
    """Deinitialize the linear stage."""
    
    def __init__(self, receiver: LinearStage150, **kwargs):
        super().__init__(receiver, **kwargs)
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize())

class LinearStage150EnableMotor(LinearStage150ParentCommand):
    """Enable the linear stage motor."""

    def __init__(self, receiver: LinearStage150, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['state'] = True
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_enabled_state(self._params['state']))
    
class LinearStage150DisableMotor(LinearStage150ParentCommand):
    """Disable the linear stage motor."""

    def __init__(self, receiver: LinearStage150, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['state'] = False
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_enabled_state(self._params['state']))

class LinearStage150MoveAbsolute(LinearStage150ParentCommand):
    """Move the linear stage to an absolute position."""

    def __init__(self, receiver: LinearStage150, position: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['position'] = position
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_absolute(self._params['position']))

class LinearStage150MoveRelative(LinearStage150ParentCommand):
    """Move the linear stage by a relative distance."""

    def __init__(self, receiver: LinearStage150, distance: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['distance'] = distance
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_relative(self._params['distance']))


from typing import List

from devices.device import Device
from .command import Command, CommandResult, CompositeCommand
from devices.keithley_2450 import Keithley2450

class KeithleyParentCommand(Command):
    """Parent class for all Keithley2450 commands."""
    receiver_cls = Keithley2450

    def __init__(self, receiver: Keithley2450, **kwargs):
        super().__init__(receiver, **kwargs)

class Keithley2450Initialize(KeithleyParentCommand):
    """Initialize the SMU by resetting it"""
    
    def __init__(self, receiver: Keithley2450, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class Keithley2450Deinitialize(KeithleyParentCommand):
    """Deinitialize the SMU"""

    #TODO: create command to deinitialize keithley2450 if necessary

    def __init__(self, reciever: Keithley2450, reset_init_flag: bool = True, **kwargs):
        super().__init__(reciever, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))


class KeithleyWait(KeithleyParentCommand):
    """Wait for all pending operations to finish"""

    def __init__(self, receiver: Keithley2450, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.wait())

class KeithleyWriteCommand(KeithleyParentCommand):
    """Write arbitrary SCPI ASCII command"""

    def __init__(self, receiver: Keithley2450, command: str, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['command'] = command

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.write_command(self._params['command']))

class KeithleySetTerminal(KeithleyParentCommand):
    """Set the terminal position of the SMU"""

    def __init__(self, receiver: Keithley2450, position: str = 'front', **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['position'] = position

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.terminal_pos(self._params['position']))

class KeithleyErrorCheck(KeithleyParentCommand):
    """Check for errors that occur during measurement"""

    def __init__(self, receiver: Keithley2450, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.error_check())

class KeithleyClearBuffer(KeithleyParentCommand):
    """Clear the data storage buffer within the Keithley2450"""
    def __init__(self, receiver: Keithley2450, buffer: str = "defbuffer1", **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['buffer'] = buffer

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.clear_buffer(self._params['buffer']))

class KeithleyIVCharacteristic(KeithleyParentCommand):
    """I-V linear sweep sourcing voltage and measuring current"""

    def __init__(self, receiver: Keithley2450, ilimit: float, vmin: float, vmax: float, delay: float, steps: int = 60, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['ilimit'] = ilimit
        self._params['vmin'] = vmin
        self._params['vmax'] = vmax
        self._params['delay'] = delay
        self._params['steps'] = steps


    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.IV_characteristic(self._params['ilimit'], self._params['vmin'], 
            self._params['vmax'], self._params['steps'], self._params['delay']))


class KeithleyFourPoint(KeithleyParentCommand):
    """Four collinear point sheet resistance measurement"""
    
    def __init__(self, receiver: Keithley2450, test_curr: float, vlimit: float, curr_reversal: bool = False, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['test_curr'] = test_curr
        self._params['vlimit'] = vlimit
        self._params['curr_reversal'] = curr_reversal

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.four_point(self._params['test_curr'], self._params['vlimit'], self._params['curr_reversal']))

class KeithleyGetData(KeithleyParentCommand):
    """Retrieve data"""

    def __init__(self, receiver: Keithley2450, filename: str = None, four_point: bool = False, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['filename'] = filename
        self._params['four_point'] = four_point

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.get_data(self._params['filename'], self._params['four_point']))





from commands.command import Command
from commands.utility_commands import LoopStartCommand, LoopEndCommand
from devices.heating_stage import HeatingStage
from devices.multi_stepper import MultiStepper
from devices.newport_esp301 import NewportESP301
from devices.festo_solenoid_valve import FestoSolenoidValve
from devices.ximea_camera import XimeaCamera
from devices.dummy_heater import DummyHeater
from devices.dummy_motor import DummyMotor
from devices.linear_stage_150 import LinearStage150
from devices.mts50_z8 import MTS50_Z8
from devices.keithley_2450 import Keithley2450
from devices.device import Device, MiscDeviceClass
from devices.utility_device import UtilityCommands

from commands.linear_stage_150_commands import *
from commands.mts50_z8_commands import *
from commands.dummy_heater_commands import *
from commands.dummy_motor_commands import *
from commands.dummy_meter_commands import *
from commands.keithley_2450_commands import *
from commands.festo_solenoid_valve_commands import *
from commands.ximea_camera_commands import *
from commands.utility_commands import *
from commands.heating_stage_commands import *
from commands.multi_stepper_commands import *
from commands.newport_esp301_commands import *
from commands.utility_commands import *

import json
import numpy as np
from typing import Tuple, Union


named_devices = {
    "PrintingStage": HeatingStage,
    "AnnealingStage": HeatingStage,
    "MultiStepper1": MultiStepper,
    "PrinterMotorX": NewportESP301,
    # "Spectrometer": StellarNetSpectrometer,
    "SampleCamera": XimeaCamera,
    "DummyHeater1": DummyHeater,
    "DummyHeater2": DummyHeater,
    "DummyMotor": DummyMotor,
    "DummyMotor1": DummyMotor,
    "DummyMotor2": DummyMotor,
}
command_directory = "commands/"
approved_devices = list(named_devices.keys())

# device_init_args = {
#     "DummyHeater": ["name", "heat_rate"],
#     "DummyMotor": ["name", "speed"],
# }


def dict_to_device(device: Device, type: str):
    device_cls = named_devices[type]
    arg_dict = device.get_init_args()

    # for attr in device_init_args[type]:
    #     arg_dict[attr] = dict["_"+attr]

    # print(arg_dict)
    return device_cls(**arg_dict)


def str_to_device(device_str: str):
    print(device_str)
    return eval(device_str)


def device_to_dict(device: Device):
    return device.get_init_args()


def evaluate(eval_str):
    return eval(eval_str)


class Encoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Device) or isinstance(obj, Command) or isinstance(obj, MiscDeviceClass):
            return obj.__dict__
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return super().default(obj)


heating_stage_ref = {
    "obj": HeatingStage,
    "serial": True,
    "serial_sequence": ["HeatingStageConnect", "HeatingStageInitialize"],
    "import_device": "from devices.heating_stage import HeatingStage",
    "import_commands": "from commands.heating_stage_commands import *",
    "init": {
        "default_code": "HeatingStage(name='Stage', port='', baudrate=115200, timeout=0.1, heating_timeout=600.0)",
        "obj_name": "HeatingStage",
        "args": {
            "name": {
                "default": "Stage",
                "type": str,
                "notes": "Name of the device",
            },
            "port": {"default": "COM", "type": str, "notes": "Port"},
            "baudrate": {
                "default": 115200,
                "type": int,
                "notes": "Baudrate",
            },
            "timeout": {
                "default": 0.1,
                "type": float,
                "notes": "Timeout",
            },
            "heating_timeout": {
                "default": 600.0,
                "type": float,
                "notes": "Heating timeout",
            },
        },
    },
    "commands": {
        "HeatingStageConnect": {
            "default_code": "HeatingStageConnect(receiver= '')",
            "args": {
                "receiver": {
                    "default": "Stage",
                    "type": str,
                    "notes": "Name of the device",
                }
            },
            "obj": HeatingStageConnect,
        },
        "HeatingStageInitialize": {
            "default_code": "HeatingStageInitialize(receiver= '')",
            "args": {
                "receiver": {
                    "default": "Stage",
                    "type": str,
                    "notes": "Name of the device",
                }
            },
            "obj": HeatingStageInitialize,
        },
        "HeatingStageDeinitialize": {
            "default_code": "HeatingStageDeinitialize(receiver= '')",
            "args": {
                "receiver": {
                    "default": "Stage",
                    "type": str,
                    "notes": "Name of the device",
                }
            },
            "obj": HeatingStageDeinitialize,
        },
        "HeatingStageSetTemp": {
            "default_code": "HeatingStageSetTemp(receiver= '', temperature= 0.0)",
            "args": {
                "receiver": {
                    "default": "Stage",
                    "type": str,
                    "notes": "Name of the device",
                },
                "temperature": {
                    "default": 0.0,
                    "type": float,
                    "notes": "Temperature",
                },
            },
            "obj": HeatingStageSetTemp,
        },
        "HeatingStageSetSetPoint": {
            "default_code": "HeatingStageSetSetPoint(receiver= '', temperature= 0.0)",
            "args": {
                "receiver": {
                    "default": "Stage",
                    "type": str,
                    "notes": "Name of the device",
                },
                "temperature": {
                    "default": 0.0,
                    "type": float,
                    "notes": "Temperature",
                },
            },
            "obj": HeatingStageSetSetPoint,
        },
    },
}


devices_ref_redundancy = {
    "UtilityCommands": {
        "obj": UtilityCommands,
        "serial": False,
        "import_device": "from devices.utility_commands import UtilityCommands",
        "import_commands": "from commands.utility_commands import *",
        "init": {
            "default_code": "# Utility Commands used",
            "obj_name": "UtilityCommands",
            "args": {},
        },
        "commands": {
            "LoopStartCommand": {
                "default_code": "LoopStartCommand()",
                "args": {},
                "obj": LoopStartCommand,
            },
            "LoopEndCommand": {
                "default_code": "LoopEndCommand()",
                "args": {},
                "obj": LoopEndCommand,
            },
            "DelayPauseCommand": {
                "default_code": "DelayPauseCommand(delay=0.0)",
                "args": {
                    "delay": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Delay in seconds",
                    }
                },
                "obj": DelayPauseCommand,
            },
            "NotifySlackCommand": {
                "default_code": "NotifySlackCommand(message='Hello World')",
                "args": {
                    "message": {
                        "default": "Hello World",
                        "type": str,
                        "notes": "Message to send to slack",
                    }
                },
                "obj": NotifySlackCommand,
            },
            "LogUserMessageCommand": {
                "default_code": "LogUserMessageCommand(message='Hello World')",
                "args": {
                    "message": {
                        "default": "Hello World",
                        "type": str,
                        "notes": "Message to log",
                    }
                },
                "obj": LogUserMessageCommand,
            },
        },
    },
    "FestoSolenoidValve": {
        "obj": FestoSolenoidValve,
        "serial": True,
        "serial_sequence": ["FestoInitialize"],
        "import_device": "from devices.festo_solenoid_valve import FestoSolenoidValve",
        "import_commands": "from commands.festo_solenoid_valve_commands import *",
        "init": {
            "default_code": "FestoSolenoidValve(name='FestoSolenoidValve', numchannel=, port='COM5', baudrate=9600, timeout=0.1)",
            "obj_name": "FestoSolenoidValve",
            "args": {
                "name": {
                    "default": "FestoSolenoidValve",
                    "type": str,
                    "notes": "Name of the device.",
                },
                "numchannel": {
                    "default": 1,
                    "type": int,
                    "notes": "",
                },
                "port": {
                    "default": "COM5",
                    "type": str,
                    "notes": "Port",
                },
                "baudrate": {
                    "default": 9600,
                    "type": int,
                    "notes": "Baudrate",
                },
                "timeout": {
                    "default": 0.1,
                    "type": float,
                    "notes": "Timeout",
                },
            },
        },
        "commands": {
            "FestoInitialize": {
                "default_code": "FestoInitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "FestoSolenoidValve",
                        "type": str,
                        "notes": "Name of the device",
                    },
                },
                "obj": FestoInitialize,
            },
            "FestoDeinitialize": {
                "default_code": "FestoDeinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "FestoSolenoidValve",
                        "type": str,
                        "notes": "Name of the device",
                    },
                },
                "obj": FestoDeinitialize,
            },
            "FestoValveOpen": {
                "default_code": "FestoValveOpen(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "FestoSolenoidValve",
                        "type": str,
                        "notes": "Name of the device",
                    },
                },
                "obj": FestoValveOpen,
            },
            "FestoValveClosed": {
                "default_code": "FestoValveClosed(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "FestoSolenoidValve",
                        "type": str,
                        "notes": "Name of the device",
                    },
                },
                "obj": FestoValveClosed,
            },
            "FestoOpenTimed": {
                "default_code": "FestoOpenTimed(receiver= '', time=0)",
                "args": {
                    "receiver": {
                        "default": "FestoSolenoidValve",
                        "type": str,
                        "notes": "Name of the device",
                    },
                    "time": {
                        "default": 0,
                        "type": int,
                        "notes": "Time to keep the valve open",
                    },
                },
                "obj": FestoOpenTimed,
            },
        },
    },
    "LinearStage150": {
        "obj": LinearStage150,
        "serial": True,
        "serial_sequence": ["LinearStage150Connect", "LinearStage150EnableMotor"],
        "import_device": "from devices.linear_stage_150 import LinearStage150",
        "import_commands": "from commands.linear_stage_150_commands import *",
        "telemetry": {
            "parameters": [
                {
                    "position": {
                        "function_name": "get_position()",
                        "data_type": "float",
                        "units": "mm",
                    }
                }
            ],
            "options": {"custom_init_args": ["port"]},
        },
        "init": {
            "default_code": "LinearStage150(name='LinearStage150', port='', baudrate=115200, timeout=0.1, destination=0x50, source=0x01, channel=1)",
            "obj_name": "LinearStage150",
            "args": {
                "name": {
                    "default": "LinearStage150",
                    "type": str,
                    "notes": "Name of the device.",
                },
                "port": {"default": "COM", "type": str, "notes": "Port"},
                "baudrate": {
                    "default": 115200,
                    "type": int,
                    "notes": "Baudrate",
                },
                "timeout": {
                    "default": 0.1,
                    "type": float,
                    "notes": "Timeout",
                },
                "destination": {
                    "default": 0x50,
                    "type": int,
                    "notes": "",
                },
                "source": {
                    "default": 0x01,
                    "type": int,
                    "notes": "",
                },
                "channel": {
                    "default": 1,
                    "type": int,
                    "notes": "",
                },
            },
        },
        "commands": {
            "LinearStage150Connect": {
                "default_code": "LinearStage150Connect(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "LinearStage150",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": LinearStage150Connect,
            },
            "LinearStage150Initialize": {
                "default_code": "LinearStage150Initialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "LinearStage150",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": LinearStage150Initialize,
            },
            "LinearStage150Deinitialize": {
                "default_code": "LinearStage150Deinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "LinearStage150",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": LinearStage150Deinitialize,
            },
            "LinearStage150EnableMotor": {
                "default_code": "LinearStage150EnableMotor(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "LinearStage150",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": LinearStage150EnableMotor,
            },
            "LinearStage150DisableMotor": {
                "default_code": "LinearStage150DisableMotor(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "LinearStage150",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": LinearStage150DisableMotor,
            },
            "LinearStage150MoveAbsolute": {
                "default_code": "LinearStage150MoveAbsolute(receiver= '', position= 0)",
                "args": {
                    "receiver": {
                        "default": "LinearStage150",
                        "type": str,
                        "notes": "",
                    },
                    "position": {
                        "default": 0,
                        "type": int,
                        "notes": "",
                    },
                },
                "obj": LinearStage150MoveAbsolute,
            },
            "LinearStage150MoveRelative": {
                "default_code": "LinearStage150MoveRelative(receiver= '', distance= 0)",
                "args": {
                    "receiver": {
                        "default": "LinearStage150",
                        "type": str,
                        "notes": "",
                    },
                    "distance": {
                        "default": 0,
                        "type": int,
                        "notes": "",
                    },
                },
                "obj": LinearStage150MoveRelative,
            },
        },
    },
    "MTS50_Z8": {
        "obj": MTS50_Z8,
        "serial": True,
        "serial_sequence": ["MTS50_Z8Connect", "MTS50_Z8EnableMotor"],
        "import_device": "from devices.mts50_z8 import MTS50_Z8",
        "import_commands": "from commands.mts50_z8_commands import *",
        "init": {
            "default_code": "MTS50_Z8(name='MTS50_Z8', port='', baudrate=115200, timeout=0.1, destination=0x50, source=0x01, channel=1)",
            "obj_name": "MTS50_Z8",
            "args": {
                "name": {
                    "default": "MTS50_Z8",
                    "type": str,
                    "notes": "Name of the device.",
                },
                "port": {"default": "COM", "type": str, "notes": "Port"},
                "baudrate": {
                    "default": 115200,
                    "type": int,
                    "notes": "Baudrate",
                },
                "timeout": {
                    "default": 0.1,
                    "type": float,
                    "notes": "Timeout",
                },
                "destination": {
                    "default": 0x50,
                    "type": int,
                    "notes": "",
                },
                "source": {
                    "default": 0x01,
                    "type": int,
                    "notes": "",
                },
                "channel": {
                    "default": 1,
                    "type": int,
                    "notes": "",
                },
            },
        },
        "commands": {
            "MTS50_Z8Connect": {
                "default_code": "MTS50_Z8Connect(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MTS50_Z8",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": MTS50_Z8Connect,
            },
            "MTS50_Z8Initialize": {
                "default_code": "MTS50_Z8Initialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MTS50_Z8",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": MTS50_Z8Initialize,
            },
            "MTS50_Z8Deinitialize": {
                "default_code": "MTS50_Z8Deinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MTS50_Z8",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": MTS50_Z8Deinitialize,
            },
            "MTS50_Z8EnableMotor": {
                "default_code": "MTS50_Z8EnableMotor(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MTS50_Z8",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": MTS50_Z8EnableMotor,
            },
            "MTS50_Z8DisableMotor": {
                "default_code": "MTS50_Z8DisableMotor(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MTS50_Z8",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": MTS50_Z8DisableMotor,
            },
            "MTS50_Z8MoveAbsolute": {
                "default_code": "MTS50_Z8MoveAbsolute(receiver= '', position= 0)",
                "args": {
                    "receiver": {
                        "default": "MTS50_Z8",
                        "type": str,
                        "notes": "",
                    },
                    "position": {
                        "default": 0,
                        "type": int,
                        "notes": "",
                    },
                },
                "obj": MTS50_Z8MoveAbsolute,
            },
            "MTS50_Z8MoveRelative": {
                "default_code": "MTS50_Z8MoveRelative(receiver= '', distance= 0)",
                "args": {
                    "receiver": {
                        "default": "MTS50_Z8",
                        "type": str,
                        "notes": "",
                    },
                    "distance": {
                        "default": 0,
                        "type": int,
                        "notes": "",
                    },
                },
                "obj": MTS50_Z8MoveRelative,
            },
        },
    },
    "Keithley2450": {
        "obj": Keithley2450,
        "serial": True,
        "serial_sequence": ["Keithley2450Initialize"],
        "import_device": "from devices.keithley_2450 import Keithley2450",
        "import_commands": "from commands.keithley_2450_commands import *",
        "init": {
            "default_code": "Keithley2450(name='Keithley2450', ID='', query_delay=0)",
            "obj_name": "Keithley2450",
            "args": {
                "name": {
                    "default": "Keithley2450",
                    "type": str,
                    "notes": "Name of the device",
                },
                "ID": {
                    "default": "",
                    "type": str,
                    "notes": "ID of the device",
                },
                "query_delay": {
                    "default": 0,
                    "type": int,
                    "notes": "Delay between queries",
                },
            },
        },
        "commands": {
            "Keithley2450Initialize": {
                "default_code": "Keithley2450Initialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": Keithley2450Initialize,
            },
            "Keithley2450Deinitialize": {
                "default_code": "Keithley2450Deinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": Keithley2450Deinitialize,
            },
            "KeithleyWait": {
                "default_code": "KeithleyWait(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": KeithleyWait,
            },
            "KeithleyWriteCommand": {
                "default_code": "KeithleyWriteCommand(receiver= '', command= '')",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    },
                    "command": {
                        "default": "",
                        "type": str,
                        "notes": "",
                    },
                },
                "obj": KeithleyWriteCommand,
            },
            "KeithleySetTerminal": {
                "default_code": "KeithleySetTerminal(receiver= '', position= 'front')",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    },
                    "position": {
                        "default": "front",
                        "type": str,
                        "notes": "",
                    },
                },
                "obj": KeithleySetTerminal,
            },
            "KeithleyErrorCheck": {
                "default_code": "KeithleyErrorCheck(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    }
                },
                "obj": KeithleyErrorCheck,
            },
            "KeithleyClearBuffer": {
                "default_code": "KeithleyClearBuffer(receiver= '', buffer='defbuffer1')",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    },
                    "buffer": {
                        "default": "defbuffer1",
                        "type": str,
                        "notes": "",
                    },
                },
                "obj": KeithleyClearBuffer,
            },
            "KeithleyIVCharacteristic": {
                "default_code": "KeithleyIVCharacteristic(receiver= '', ilimit=0.0, vmin=0.0, vmax=0.0, delay=0.0, steps=60)",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    },
                    "ilimit": {
                        "default": 0.0,
                        "type": float,
                        "notes": "",
                    },
                    "vmin": {
                        "default": 0.0,
                        "type": float,
                        "notes": "",
                    },
                    "vmax": {
                        "default": 0.0,
                        "type": float,
                        "notes": "",
                    },
                    "delay": {
                        "default": 0.0,
                        "type": float,
                        "notes": "",
                    },
                    "steps": {
                        "default": 60,
                        "type": int,
                        "notes": "",
                    },
                },
                "obj": KeithleyIVCharacteristic,
            },
            "KeithleyFourPoint": {
                "default_code": "KeithleyFourPoint(receiver= '', test_curr=0.0, vlimit=0.0, curr_reversal = False)",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    },
                    "test_curr": {
                        "default": 0.0,
                        "type": float,
                        "notes": "",
                    },
                    "vlimit": {
                        "default": 0.0,
                        "type": float,
                        "notes": "",
                    },
                    "curr_reversal": {
                        "default": False,
                        "type": bool,
                        "notes": "",
                    },
                },
                "obj": KeithleyFourPoint,
            },
            "KeithleyGetData": {
                "default_code": "KeithleyGetData(receiver= '', filename=None, four_point= False)",
                "args": {
                    "receiver": {
                        "default": "Keithley2450",
                        "type": str,
                        "notes": "",
                    },
                    "filename": {
                        "default": None,
                        "type": str,
                        "notes": "",
                    },
                    "four_point": {
                        "default": False,
                        "type": bool,
                        "notes": "",
                    },
                },
                "obj": KeithleyGetData,
            },
        },
    },
    "PrintingStage": heating_stage_ref,
    "AnnealingStage": heating_stage_ref,
    "MultiStepper": {
        "obj": MultiStepper,
        "serial": True,
        "serial_sequence": ["MultiStepperConnect", "MultiStepperInitialize"],
        "import_device": "from devices.multi_stepper import MultiStepper",
        "import_commands": "from commands.multi_stepper_commands import *",
        "init": {
            "default_code": "MultiStepper(name='MultiStepper', port='', baudrate=115200, timeout=0.1, destination=0x50, source=0x01, channel=1)",
            "obj_name": "MultiStepper",
            "args": {
                "name": {
                    "default": "MultiStepper",
                    "type": str,
                    "notes": "Name of the device",
                },
                "port": {
                    "default": "COM",
                    "type": str,
                    "notes": "Port of the device",
                },
                "baudrate": {
                    "default": 115200,
                    "type": int,
                    "notes": "Baudrate of the device",
                },
                "timeout": {
                    "default": 1.0,
                    "type": float,
                    "notes": "Timeout of the device",
                },
                "stepper_list": {
                    "default": (1,),
                    "type": Tuple[int, ...],
                    "notes": "List of stepper numbers",
                },
                "move_timeout": {
                    "default": 30.0,
                    "type": float,
                    "notes": "Timeout for move commands",
                },
            },
        },
        "commands": {
            "MultiStepperConnect": {
                "default_code": "MultiStepperConnect(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MultiStepper",
                        "type": str,
                        "notes": "Name of the device",
                    }
                },
                "obj": MultiStepperConnect,
            },
            "MultiStepperInitialize": {
                "default_code": "MultiStepperInitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MultiStepper",
                        "type": str,
                        "notes": "Name of the device",
                    }
                },
                "obj": MultiStepperInitialize,
            },
            "MultiStepperDeinitialize": {
                "default_code": "MultiStepperDeinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "MultiStepper",
                        "type": str,
                        "notes": "Name of the device",
                    }
                },
                "obj": MultiStepperDeinitialize,
            },
            "MultiStepperMoveAbsolute": {
                "default_code": "MultiStepperMoveAbsolute(receiver= '', stepper_number= 0, position= 0)",
                "args": {
                    "receiver": {
                        "default": "MultiStepper",
                        "type": str,
                        "notes": "Name of the device",
                    },
                    "stepper_number": {
                        "default": 0,
                        "type": int,
                        "notes": "Number of the stepper",
                    },
                    "position": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Position to move to",
                    },
                },
                "obj": MultiStepperMoveAbsolute,
            },
            "MultiStepperMoveRelative": {
                "default_code": "MultiStepperMoveRelative(receiver= '', stepper_number= 0, distance= 0)",
                "args": {
                    "receiver": {
                        "default": "MultiStepper",
                        "type": str,
                        "notes": "Name of the device",
                    },
                    "stepper_number": {
                        "default": 0,
                        "type": int,
                        "notes": "Number of the stepper",
                    },
                    "distance": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Distance to move",
                    },
                },
                "obj": MultiStepperMoveRelative,
            },
        },
    },
    "NewportESP301": {
        "obj": NewportESP301,
        "serial": True,
        "serial_sequence": ["NewportESP301Connect", "NewportESP301Initialize"],
        "import_device": "from devices.newport_esp301 import NewportESP301",
        "import_commands": "from commands.newport_esp301_commands import *",
        "init": {
            "default_code": "NewportESP301(name='NewportESP301', port='', baudrate=921600, timeout=1.0, axis_list = (1,), default_speed=20.0, poll_interval=0.1)",
            "obj_name": "NewportESP301",
            "args": {
                "name": {
                    "default": "NewportESP301",
                    "type": str,
                    "notes": "Name of the device",
                },
                "port": {
                    "default": "COM",
                    "type": str,
                    "notes": "Port of the device",
                },
                "baudrate": {
                    "default": 921600,
                    "type": int,
                    "notes": "Baudrate of the device",
                },
                "timeout": {
                    "default": 1.0,
                    "type": float,
                    "notes": "Timeout of the device",
                },
                "axis_list": {
                    "default": (1,),
                    "type": Tuple[int, ...],
                    "notes": "List of axis numbers",
                },
                "default_speed": {
                    "default": 20.0,
                    "type": float,
                    "notes": "Default speed of the device",
                },
                "poll_interval": {
                    "default": 0.1,
                    "type": float,
                    "notes": "Poll interval of the device",
                },
            },
        },
        "commands": {
            "NewportESP301Connect": {
                "default_code": "NewportESP301Connect(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "NewportESP301",
                        "type": str,
                        "notes": "Name of the device",
                    }
                },
                "obj": NewportESP301Connect,
            },
            "NewportESP301Initialize": {
                "default_code": "NewportESP301Initialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "NewportESP301",
                        "type": str,
                        "notes": "Name of the device",
                    }
                },
                "obj": NewportESP301Initialize,
            },
            "NewportESP301Deinitialize": {
                "default_code": "NewportESP301Deinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "NewportESP301",
                        "type": str,
                        "notes": "Name of the device",
                    }
                },
                "obj": NewportESP301Deinitialize,
            },
            "NewportESP301MoveSpeedAbsolute": {
                "default_code": "NewportESP301MoveSpeedAbsolute(receiver= '', axis= 1, position= 0, speed= 20.0)",
                "args": {
                    "receiver": {
                        "default": "NewportESP301",
                        "type": str,
                        "notes": "Name of the device",
                    },
                    "axis": {
                        "default": 1,
                        "type": int,
                        "notes": "Axis number",
                    },
                    "position": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Position to move to",
                    },
                    "speed": {
                        "default": 20.0,
                        "type": float,
                        "notes": "Speed to move at",
                    },
                },
                "obj": NewportESP301MoveSpeedAbsolute,
            },
            "NewportESP301MoveSpeedRelative": {
                "default_code": "NewportESP301MoveSpeedRelative(receiver= '', axis= 1, distance= 0, speed= 20.0)",
                "args": {
                    "receiver": {
                        "default": "NewportESP301",
                        "type": str,
                        "notes": "Name of the device",
                    },
                    "axis": {
                        "default": 1,
                        "type": int,
                        "notes": "Axis number",
                    },
                    "distance": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Distance to move",
                    },
                    "speed": {
                        "default": 20.0,
                        "type": float,
                        "notes": "Speed to move at",
                    },
                },
                "obj": NewportESP301MoveSpeedRelative,
            },
        },
    },
    "DummyMotor": {
        "obj": DummyMotor,
        "serial": True,
        "serial_sequence": ["DummyMotorInitialize"],
        "import_device": "from devices.dummy_motor import DummyMotor",
        "import_commands": "from commands.dummy_motor_commands import *",
        "init": {
            "default_code": "DummyMotor(name='DummyMotor', speed=20.0)",
            "obj_name": "DummyMotor",
            "args": {
                "name": {
                    "default": "DummyMotor",
                    "type": str,
                    "notes": "Name of the device.",
                },
                "speed": {
                    "default": 20.0,
                    "type": float,
                    "notes": "Speed of the motor.",
                },
            },
        },
        "commands": {
            "DummyMotorInitialize": {
                "default_code": "DummyMotorInitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "DummyMotor",
                        "type": str,
                        "notes": "Name of the device.",
                    }
                },
                "obj": DummyMotorInitialize,
            },
            "DummyMotorDeinitialize": {
                "default_code": "DummyMotorDeinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "DummyMotor",
                        "type": str,
                        "notes": "Name of the device.",
                    }
                },
                "obj": DummyMotorDeinitialize,
            },
            "DummyMotorSetSpeed": {
                "default_code": "DummyMotorSetSpeed(receiver= '', speed= 0.0)",
                "args": {
                    "receiver": {
                        "default": "DummyMotor",
                        "type": str,
                        "notes": "Name of the device.",
                    },
                    "speed": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Speed of the motor.",
                    },
                },
                "obj": DummyMotorSetSpeed,
            },
            "DummyMotorMoveAbsolute": {
                "default_code": "DummyMotorMoveAbsolute(receiver= '', position= 0)",
                "args": {
                    "receiver": {
                        "default": "DummyMotor",
                        "type": str,
                        "notes": "Name of the device.",
                    },
                    "position": {
                        "default": 0,
                        "type": float,
                        "notes": "Position to move to.",
                    },
                },
                "obj": DummyMotorMoveAbsolute,
            },
            "DummyMotorMoveRelative": {
                "default_code": "DummyMotorMoveRelative(receiver= '', distance= 0)",
                "args": {
                    "receiver": {
                        "default": "DummyMotor",
                        "type": str,
                        "notes": "Name of the device.",
                    },
                    "distance": {
                        "default": 0,
                        "type": float,
                        "notes": "Distance to move.",
                    },
                },
                "obj": DummyMotorMoveRelative,
            },
            "DummyMotorMoveSpeedAbsolute": {
                "default_code": "DummyMotorMoveSpeedAbsolute(receiver= '', position= 0.0, speed= 0.0)",
                "args": {
                    "receiver": {
                        "default": "DummyMotor",
                        "type": str,
                        "notes": "Name of the device.",
                    },
                    "position": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Position to move to.",
                    },
                    "speed": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Speed of the motor.",
                    },
                },
                "obj": DummyMotorMoveSpeedAbsolute,
            },
        },
    },
    # "Spectrometer": {"obj": StellarNetSpectrometer},
    # "XimeaCamera": {"obj": XimeaCamera},
    "DummyHeater": {
        "obj": DummyHeater,
        "serial": True,
        "serial_sequence": ["DummyHeaterInitialize"],
        "import_device": "from devices.dummy_heater import DummyHeater",
        "import_commands": "from commands.dummy_heater_commands import *",
        "init": {
            "default_code": "DummyHeater(name='DummyHeater', heat_rate=20.0)",
            "obj_name": "DummyHeater",
            "args": {
                "name": {
                    "default": "DummyHeater",
                    "type": str,
                    "notes": "Name of the device.",
                },
                "heat_rate": {
                    "default": 20.0,
                    "type": float,
                    "notes": "Heat rate of the heater.",
                },
            },
        },
        "commands": {
            "DummyHeaterInitialize": {
                "default_code": "DummyHeaterInitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "DummyHeater",
                        "type": str,
                        "notes": "Name of the device.",
                    }
                },
                "obj": DummyHeaterInitialize,
            },
            "DummyHeaterDeinitialize": {
                "default_code": "DummyHeaterDeinitialize(receiver= '')",
                "args": {
                    "receiver": {
                        "default": "DummyHeater",
                        "type": str,
                        "notes": "Name of the device.",
                    }
                },
                "obj": DummyHeaterDeinitialize,
            },
            "DummyHeaterSetHeatRate": {
                "default_code": "DummyHeaterSetHeatRate(receiver= '', heat_rate= 0.0)",
                "args": {
                    "receiver": {
                        "default": "DummyHeater",
                        "type": str,
                        "notes": "Name of the device.",
                    },
                    "heat_rate": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Heat rate of the heater.",
                    },
                },
                "obj": DummyHeaterSetHeatRate,
            },
            "DummyHeaterSetTemp": {
                "default_code": "DummyHeaterSetTemp(receiver= '', temperature= 0.0)",
                "args": {
                    "receiver": {
                        "default": "DummyHeater",
                        "type": str,
                        "notes": "Name of the device.",
                    },
                    "temperature": {
                        "default": 0.0,
                        "type": float,
                        "notes": "Temperature to set the heater to.",
                    },
                },
                "obj": DummyHeaterSetTemp,
            },
        },
    },
}


# devices_ref = {
#     "PrintingStage": heating_stage_ref,
#     "AnnealingStage": heating_stage_ref,
#     "MultiStepper": {
#         "obj": MultiStepper,
#         "import_device": "from devices.multi_stepper import MultiStepper",
#         "import_commands": "from commands.multi_stepper_commands import *",
#         "init": "MultiStepper(name='MultiStepper', port='', baudrate=115200, timeout=0.1, destination=0x50, source=0x01, channel=1)",
#         "commands": {
#             "MultiStepperConnect": "MultiStepperConnect(receiver= '')",
#             "MultiStepperInitialize": "MultiStepperInitialize(receiver= '')",
#             "MultiStepperDeinitialize": "MultiStepperDeinitialize(receiver= '')",
#             "MultiStepperMoveAbsolute": "MultiStepperMoveAbsolute(receiver= '', stepper_number= 0, position= 0)",
#             "MultiStepperMoveRelative": "MultiStepperMoveRelative(receiver= '', stepper_number= 0, distance= 0)",
#         },
#     },
#     "PrinterMotorX": {"obj": NewportESP301},
#     # "Spectrometer": {"obj": StellarNetSpectrometer},
#     "XimeaCamera": {"obj": XimeaCamera},
#     "DummyHeater": {"obj": DummyHeater},
#     "DummyMotor": {"obj": DummyMotor},
#     "LinearStage150": {
#         "obj": LinearStage150,
#         "import_device": "from devices.linear_stage_150 import LinearStage150",
#         "import_commands": "from commands.linear_stage_150_commands import *",
#         "init": "LinearStage150(name='LinearStage150', port='', baudrate=115200, timeout=0.1, destination=0x50, source=0x01, channel=1)",
#         "commands": {
#             "LinearStage150Connect": "LinearStage150Connect(receiver= '')",
#             "LinearStage150Initialize": "LinearStage150Initialize(receiver= '')",
#             "LinearStage150Deinitialize": "LinearStage150Deinitialize(receiver= '')",
#             "LinearStage150EnableMotor": "LinearStage150EnableMotor(receiver= '')",
#             "LinearStage150DisableMotor": "LinearStage150DisableMotor(receiver= '')",
#             "LinearStage150MoveAbsolute": "LinearStage150MoveAbsolute(receiver= '', position= 0)",
#             "LinearStage150MoveRelative": "LinearStage150MoveRelative(receiver= '', distance= 0)",
#         },
#     },
# }



import time
from typing import Tuple

from .device import Device, check_initialized
from .dummy_motor_source import DummyMotorSource


class DummyMotor(Device):
    def __init__(self, name: str, speed: float = 20.0):
        super().__init__(name)
        # Want to inherit from Device to implement standard name and initialization attributes/methods
        # Using composition instead of multiple inheritance
        self.motor = DummyMotorSource(speed)

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self._name,
            "speed": self.motor._speed,
        }
        return args_dict
    
    def update_init_args(self, args_dict: dict):
        self.motor._speed = args_dict["speed"]
        self._name = args_dict["name"]

    @property
    def position(self) -> float:
        return self.motor.position
    
    @property
    def speed(self) -> float:
        return self.motor.speed

    def initialize(self) -> Tuple[bool, str]:
        self.motor.home_motor()
        while self.motor._position != 0.0:
            print("DummyMotor " + self.name + " position: " + str(self.motor._position), end='\r')
            time.sleep(0.1)
        print("DummyMotor " + self.name + " position: " + str(self.motor._position), end='\r')
        self._is_initialized = True
        return (True, "Initialized DummyMotor by homing and setting position to zero")

    def deinitialize(self, reset_init_flag: bool = True) -> Tuple[bool, str]:
        self.move_absolute(position=0.0)
        if reset_init_flag:
            self._is_initialized = False
        return (True, "Deinitialized DummyMotor by moving to position zero")

    def set_speed(self, speed: float):
        self.motor.speed = speed
        # if out of range, then the setter did nothing
        if self.motor.speed == speed:
            return (True, "DummyMotor speed was successfully set to " + str(self.motor.speed))
        else:
            return (False, "DummyMotor speed was not set. Speed is currently " + str(self.motor.speed))

    @check_initialized
    def move_absolute(self, position: float) -> Tuple[bool, str]:
        # if not self._is_initialized:
        #     return (False, "DummyMotor is not initialized")
        if not self.is_valid_position(position):
            return (False, "Position is not valid")

        self.motor.move_absolute(position)

        while self.motor.position != position:
            print("DummyMotor " + self.name + " position: " + str(self.motor.position), end='\r')
            time.sleep(.1)
        print("DummyMotor " + self.name + " position: " + str(self.motor.position), end='\r')

        return (True, "DummyMotor has reached position " + str(position))

    @check_initialized
    def move_relative(self, distance: float) -> Tuple[bool, str]:
        # if not self._is_initialized:
        #     return (False, "DummyMotor is not initialized")

        position = self.motor.position + distance
        if not self.is_valid_position(position):
            return (False, "Position is not valid")

        self.motor.move_relative(distance)

        while self.motor.position != position:
            print("DummyMotor " + self.name + " position: " + str(self.motor.position), end='\r')
            time.sleep(.1)
        print("DummyMotor " + self.name + " position: " + str(self.motor.position), end='\r')

        return (True, "DummyMotor has moved by " + str(distance) + " and reached position " + str(position))

    def is_valid_position(self, position) -> bool:
        if position >= self.motor.min_position and position <= self.motor.max_position:
            return True
        else:
            return False

    



from typing import Optional, Tuple, Union
import serial

from .device import ArduinoSerialDevice, check_initialized, check_serial


class HeatingStage(ArduinoSerialDevice):
    def __init__(
            self, 
            name: str,  
            port: str, 
            baudrate: int, 
            timeout: Optional[float] = 1.0, 
            heating_timeout: float = 600.0):

        super().__init__(name, port, baudrate, timeout)
        self._heating_timeout = heating_timeout

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self.name,
            "port": self._port,
            "baudrate": self._baudrate,
            "timeout": self._timeout,
            "heating_timeout": self._heating_timeout,
        }
        return args_dict
    
    def update_init_args(self, args_dict: dict):
        self.name = args_dict["name"]
        self._port = args_dict["port"]
        self._baudrate = args_dict["baudrate"]
        self._timeout = args_dict["timeout"]
        self._heating_timeout = args_dict["heating_timeout"]

    # no need to check serial as set_settemp and pid_on has these checks already
    def initialize(self) -> Tuple[bool, str]:
        self._is_initialized = True
        was_set, comment = self.set_settemp(26.0)

        if not was_set:
            self._is_initialized = False
            return (was_set, comment)

        was_turned_on, comment = self.pid_on()

        if not was_turned_on:
            self._is_initialized = False
            return (was_turned_on, comment)

        return (True, "Heating stage successfully initialized by setting to 26 C and turning PID ON.")

    def deinitialize(self, reset_init_flag: bool = True) -> Tuple[bool, str]:
        self._is_initialized = True
        was_set, comment = self.set_settemp(24.0)

        if not was_set:
            self._is_initialized = False
            return (was_set, comment)

        was_turned_off, comment = self.pid_off()

        if not was_turned_off:
            self._is_initialized = False
            return (was_turned_off, comment)

        if reset_init_flag:
            self._is_initialized = False

        return (True, "Heating stage successfully deinitialized by setting to 24 C and turning PID OFF.")

    @check_serial
    @check_initialized
    def set_settemp(self, temp: float) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # elif not self._is_initialized:
        #     return (False, "Heating stage is not initialized.")
        # else:
        command = ">set Ts " + str(temp) +"\n"
        self.ser.write(command.encode('ascii'))
        return self.check_ack_succ()

    @check_serial
    @check_initialized
    def set_temp(self, temp: float) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # elif not self._is_initialized:
        #     return (False, "Heating stage is not initialized.")
        # else:
        command = ">set T " + str(temp) +"\n"
        self.ser.write(command.encode('ascii'))
        return self.check_ack_succ(ack_timeout=1.0, succ_timeout=self._heating_timeout)

    @check_serial
    @check_initialized
    def pid_on(self) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # elif not self._is_initialized:
        #     return (False, "Heating stage is not initialized.")
        # else:
        command = ">pidon\n"
        self.ser.write(command.encode('ascii'))
        return self.check_ack_succ()

    @check_serial
    def pid_off(self) -> Tuple[bool, str]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # else:
        command = ">pidoff\n"
        self.ser.write(command.encode('ascii'))
        return self.check_ack_succ()
    
    @check_serial
    def is_pid_on(self) -> Tuple[bool, Union[bool, str]]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # else:
        command = ">pr pid\n"
        self.ser.write(command.encode('ascii'))
        was_successful, comment = self.check_ack_succ()

        if not was_successful:
            return (was_successful, comment)

        has_on_off, pid_state = super().parse_equal_sign(comment)

        if not has_on_off:
            return (has_on_off, "Message from device did not contain PID state.")

        if pid_state == "ON":
            is_on = True
        else:
            is_on = False

        return (True, is_on)

    @check_serial
    def temperature(self) -> Tuple[bool, Union[float, str]]:
        # if not self.ser.is_open:
        #     return (False, "Serial port " + self._port + " is not open. ")
        # else:
        command = ">pr T\n"
        self.ser.write(command.encode('ascii'))
        was_successful, comment = self.check_ack_succ()

        if not was_successful:
            return (was_successful, comment)

        has_temperature, temperature_str = super().parse_equal_sign(comment)

        if not has_temperature:
            return (has_temperature, "Message from device did not contain temperature.")
        
        return (True, float(temperature_str))



    



import threading
import time
from typing import Tuple
import random

from .device import Device, check_initialized

class DummyHeater(Device):

    def __init__(self, name: str, heat_rate: float = 20.0):
        super().__init__(name)
        self._heat_rate = heat_rate
        self.min_heat_rate = 1.0
        self.max_heat_rate = 50.0
        self.min_temperature = 25.0
        self.max_temperature = 100.0
        self._temperature = random.uniform(self.min_temperature, self.max_temperature)
        self._hardware_interval = 0.05

    def get_init_args(self) -> dict:
        args_dict = {
            "name": self._name,
            "heat_rate": self._heat_rate,
        }
        return args_dict
    
    def update_init_args(self, args_dict: dict):
        self._name = args_dict["name"]
        self._heat_rate = args_dict["heat_rate"]


    @property
    def temperature(self) -> float:
        return self._temperature

    @property
    def heat_rate(self) -> float:
        return self._heat_rate

    @heat_rate.setter
    def heat_rate(self, heat_rate: float):
        if heat_rate >= self.min_heat_rate and heat_rate <= self.max_heat_rate:
            self._heat_rate = heat_rate
        # The logic for success bool and result message is implemented in the SetHeatRate command
    
    def initialize(self) -> Tuple[bool, str]:
        self._is_initialized = True # this would normally go at the end of this method, but it needs to be true to set temp
        self.set_temp(25.0)
        return (True, "Initialized DummyHeater by setting to room temperature (25 C)")

    def deinitialize(self, reset_init_flag: bool = True) -> Tuple[bool, str]:
        self.set_temp(25.0)
        if reset_init_flag:
            self._is_initialized = False
        return (True, "Deinitialized DummyHeater by setting to room temperature (25 C)")

    # set_temp waits for "hardware" to reach temperature set point before returning
    @check_initialized
    def set_temp(self, temperature: float) -> Tuple[bool, str]:
        # if not self._is_initialized:
        #     return (False, "DummyHeater is not initialized")
        if not self.is_valid_temp(temperature):
            return (False, "Temperature setpoint is invalid")

        temp_diff = temperature - self._temperature
        if temp_diff > 0.0:
            self._hardware_heat(temperature, 1)
        elif temp_diff < 0.0:
            self._hardware_heat(temperature, -1)

        while self._temperature != temperature:
            print("DummyHeater " + self.name + " temperature: " + str(self._temperature), end='\r')
            time.sleep(.1)
        print("DummyHeater " + self.name + " temperature: " + str(self._temperature), end='\r')

        return (True, "DummyHeater temperature has stabilized at " + str(temperature))

    def is_valid_temp(self, temperature: float) -> bool:
        if temperature >= self.min_temperature and temperature <= self.max_temperature:
            return True
        else:
            return False

    # Emulates constant temperature change
    def _hardware_heat(self, target: float, direction: int):
        # direction is 1 or -1
        timer = threading.Timer(self._hardware_interval, self._hardware_heat, [target, direction])
        timer.start()
        temperature_change = self._heat_rate * self._hardware_interval * direction

        if abs(target - self._temperature) <= abs(temperature_change):
            self._temperature = target
            timer.cancel()
        else:
            self._temperature += temperature_change


from typing import Optional

from .command import Command, CommandResult
from devices.ximea_camera import XimeaCamera


class XimeaCameraParentCommand(Command):
    """Parent class for all XimeaCamera commands."""
    receiver_cls = XimeaCamera

    def __init__(self, receiver: XimeaCamera, **kwargs):
        super().__init__(receiver, **kwargs)

class XimeaCameraInitialize(XimeaCameraParentCommand):
    """Initialize camera by opening communication with camera and optionally set defaults."""

    def __init__(self, receiver: XimeaCamera, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class XimeaCameraDeinitialize(XimeaCameraParentCommand):
    """Deinitialize camera by stopping any acquisition and closing communication with camera."""

    def __init__(self, receiver: XimeaCamera, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))

class XimeaCameraGetImage(XimeaCameraParentCommand):
    """Get image and save to file, display image, or both. No filename = timestamped filename. No exposure or gain = use defaults."""
    
    def __init__(
            self, 
            receiver: XimeaCamera, 
            save_to_file: bool = True, 
            filename: str = None, 
            exposure_time: Optional[int] = None, 
            gain: Optional[float] = None, 
            show_pop_up: bool = False, 
            **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['save_to_file'] = save_to_file
        self._params['filename'] = filename
        self._params['exposure_time'] = exposure_time
        self._params['gain'] = gain
        self._params['show_pop_up'] = show_pop_up

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.get_image(self._params['save_to_file'], self._params['filename'], self._params['exposure_time'], self._params['gain'], self._params['show_pop_up']))

class XimeaCameraSetDefaultExposure(XimeaCameraParentCommand):
    """Set the default exposure time to use if no exposure time is passed when getting image."""

    def __init__(self, receiver: XimeaCamera, exposure_time: int, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['exposure_time'] = exposure_time

    def execute(self) -> None:
        self._receiver.default_exposure_time = self._params['exposure_time']
        if self._receiver.default_exposure_time == self._params['exposure_time']:
            self._result = CommandResult(True, "Default exposure time successfully set to " + str(self._params['exposure_time']))
        else:
            self._result = CommandResult(False, "Failed to set the default exposure time to " + str(self._params['exposure_time']) + ". The default exposure time must be > 0.")

class XimeaCameraSetDefaultGain(XimeaCameraParentCommand):
    """Set the default gain to use if no gain is passed when getting image."""

    def __init__(self, receiver: XimeaCamera, gain: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['gain'] = gain

    def execute(self) -> None:
        self._receiver.default_gain = self._params['gain']
        if self._receiver.default_gain == self._params['gain']:
            self._result = CommandResult(True, "Default gain successfully set to " + str(self._params['gain']))
        else:
            # unsure what the upper limit is at the moment
            self._result = CommandResult(False, "Failed to set the default gain to " + str(self._params['gain']) + ". The default gain must be >= 0 and has an upper limit.")

class XimeaCameraUpdateWhiteBal(XimeaCameraParentCommand):
    """Update the white balance coefficients using the current camera feed."""

    def __init__(self, receiver: XimeaCamera, exposure_time: int = None, gain: Optional[float] = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['exposure_time'] = exposure_time
        self._params['gain'] = gain

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.update_white_balance(self._params['exposure_time'], self._params['gain']))

class XimeaCameraSetManualWhiteBal(XimeaCameraParentCommand):
    """Set any of the red, green, and blue white balance coefficients manually."""

    def __init__(self, receiver: XimeaCamera, wb_kr: Optional[float] = None, wb_kg: Optional[float] = None, wb_kb: Optional[float] = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['wb_kr'] = wb_kr
        self._params['wb_kg'] = wb_kg
        self._params['wb_kb'] = wb_kb
    
    def execute(self) -> None:
        self._result = CommandResult(self._receiver.set_white_balance_manually(self._params['wb_kr'], self._params['wb_kg'], self._params['wb_kb']))

class XimeaCameraResetWhiteBal(XimeaCameraParentCommand):
    """Reset the red, green, and blue white balance coefficients to defaults."""

    def __init__(self, receiver: XimeaCamera, **kwargs):
        super().__init__(receiver, **kwargs)
    
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.reset_white_balance_rgb_coeffs())




from .device import Device

class UtilityCommands(Device):
    def __init__(self, name: str):
        super().__init__(name)

    def get_init_args(self) -> dict:
        return {"name": self._name}
    
    def update_init_args(self, args_dict: dict):
        self._name = args_dict["name"]


from .command import Command, CommandResult

class GenericInitialize(Command):
    """desc"""

    def __init__(self, receiver, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        try:
            self._result = CommandResult(*self._receiver.initialize())
        except AttributeError as err:
            self._result = CommandResult(False, str(err))

class GenericDeinitialize(Command):
    """desc"""

    def __init__(self, receiver, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        try:
            self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))
        except AttributeError as err:
            self._result = CommandResult(False, str(err))

class GenericSerialConnect(Command):
    """desc"""

    def __init__(self, receiver, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        try:
            self._result = CommandResult(*self._receiver.start_serial())
        except AttributeError as err:
            self._result = CommandResult(False, str(err))

    


from .command import Command, CommandResult
from devices.psd6_syringe_pump import PSD6SyringePump
from typing import Optional

# Parent class, subclass from Command ABC
class PSD6SyringePumpParentCommand(Command):
    """Parent class for all PSD6SyringePump commands."""
    receiver_cls = PSD6SyringePump

    def __init__(self, receiver: PSD6SyringePump, **kwargs):
        super().__init__(receiver, **kwargs)

# Recommended command classes
class PSD6SyringePumpConnect(PSD6SyringePumpParentCommand):
    """Open the serial port to the PSD6 syringe pump."""

    def __init__(self, receiver: PSD6SyringePump, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.start_serial())
        
class PSD6SyringePumpInitialize(PSD6SyringePumpParentCommand):
    def __init__(self, receiver: PSD6SyringePump, **kwargs):
        super().__init__(receiver, **kwargs)
        
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())
        
class PSD6SyringePumpMoveValve(PSD6SyringePumpParentCommand):
    def __init__(self, receiver: PSD6SyringePump, valve_num: int, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['valve_num'] = valve_num
        
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_valve_position(self._params['valve_num']))
        
class PSD6SyringePumpMoveAbsolute(PSD6SyringePumpParentCommand):
    def __init__(self, receiver: PSD6SyringePump, volume: float, valve_num: Optional[int] = None, flowrate: Optional[float] = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['volume'] = volume
        self._params['valve_num'] = valve_num
        self._params['flowrate'] = flowrate
        
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_syringe_absolute_volume(self._params['volume'], self._params['valve_num'], self._params['flowrate']))
        
class PSD6SyringePumpInfuse(PSD6SyringePumpParentCommand):
    def __init__(self, receiver: PSD6SyringePump, volume: float, valve_num: Optional[int] = None, flowrate: Optional[float] = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['volume'] = volume
        self._params['valve_num'] = valve_num
        self._params['flowrate'] = flowrate
        
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.infuse_syringe_volume(self._params['volume'], self._params['valve_num'], self._params['flowrate']))

class PSD6SyringePumpWithdraw(PSD6SyringePumpParentCommand):
    def __init__(self, receiver: PSD6SyringePump, volume: float, valve_num: Optional[int] = None, flowrate: Optional[float] = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['volume'] = volume
        self._params['valve_num'] = valve_num
        self._params['flowrate'] = flowrate
        
    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.withdraw_syringe_volume(self._params['volume'], self._params['valve_num'], self._params['flowrate']))


from abc import ABC, abstractmethod
from typing import Optional
import time

from devices.device import Device



# TODO
# logging from within the composite command execution?
# The adding of args into params dict can be done with locals()

class Command(ABC):
    """The Command abstract base class which acts as an interface for other objects that use commands."""

    receiver_cls = None # or Device?

    @abstractmethod
    def __init__(self, receiver: Device, delay: float = 0.0):
        # Child classes will still have 'receiver' in their signature (separate from **kwargs) because I want the IDE to hint at the specific receiver class
        # delay will be passed up through **kwargs as it is an secondary parameter for all commands 
        self._receiver = receiver
        self._params = {}
        self._params['receiver_name'] = receiver.name
        self._params['delay'] = delay
        # self._was_successful = None
        # self._result_message = None  

        # this could also just be None, instead of a CommandResult with None attributes  
        # is there ever a time where the result would be accessed before the command is executed?
        # it might be better to just have it as a CommandResult with None's since whatever accessing it
        # is expecting _result to be a CommandResult object anyways
        # but its possible even after successful execution that some attributes are still None
        # Therefore there is a difference between whether an accessed attribute is None or if the entire result object is None
        self._result = CommandResult()

    # definitely enforce return None for execute()
    @abstractmethod
    def execute(self) -> None:
        """Execute abstract method from which all commands implement; accepts no arguments and returns None."""
        pass

    @property
    def name(self) -> str:
        """The command's name getter.

        Returns
        -------
        str
            Returns the command's class name followed by its parameters.
        """
        self._name = type(self).__name__
        for key, value in self._params.items():
            if not key == 'delay':
                self._name += " " + key + "=" + str(value)
        # I want delay to always be last when displayed and only if its not zero
        if isinstance(self._params['delay'], str):
            self._name += " " + 'delay=' + str(self._params['delay'])
        else:
            if self._params['delay'] > 0.0:
                self._name += " " + 'delay=' + str(self._params['delay'])
        return self._name

    @property
    def description(self) -> str:
        """The command's description getter.

        Returns
        -------
        str
            Returns the command's docstring
        """
        return self.__doc__

    # params and receiver getter?

    # type hint return CustomResult?
    @property
    def result(self):
        return self._result

    # @property
    # def was_successful(self) -> Optional[bool]:
    #     """Whether the command's execution was successful or not.

    #     Returns
    #     -------
    #     Optional[bool]
    #         Returns True if the command's execution was successful, otherwise False. Initialized with None.
    #     """
    #     return self._was_successful

    # @property
    # def result_message(self) -> Optional[str]:
    #     """A message describing the result of the command's execution

    #     Returns
    #     -------
    #     Optional[str]
    #         Returns a string describing the result of the command's execution with details if it failed.
    #     """
    #     return self._result_message

    def get_init_args(self) -> dict:
        """Get the command's initialization arguments.

        Returns
        -------
        dict
            Returns a dictionary of the command's initialization arguments.
        """
        return self._params

# store info like name and description in here too? or leave separate? Leave separate, semantically i think it makes sense 
# to retrive non-result info from the command instead of the commandresult object, potentially avoid conflicting info too
# Mainly adding to future proof in case more info needs to be retrieved after execution
class CommandResult():
    """Object which stores whether a command's execution was successful and other relevant information"""

    def __init__(self, was_successful: Optional[bool] = None, message: Optional[str] = None):
        # I considered whether to have a reset/update function instead and have init call it, this function could be used outside to reset/update the result attributes
        # The alternative is to just use the contructor to create a new reset object. I decided just using the constructor is better.
        # Although it creates a new object in memory, it ensures that whenever a command executes if it uses the constructor then the results 
        # will be fully reset and not hold any old info if it is executed multiples times. In other cases, it may hold old info like for 
        # 1) explicitly tuple unpacking in the result attributes (old version) or 2) using an update function (which may not update all attrs)
        
        self._was_successful = was_successful
        self._message = message
    
    @property
    def was_successful(self) -> Optional[bool]:
        """Whether the command's execution was successful or not.

        Returns
        -------
        Optional[bool]
            Returns True if the command's execution was successful, otherwise False. Initialized with None.
        """
        return self._was_successful

    @property
    def message(self) -> Optional[str]:
        """A message describing the result of the command's execution

        Returns
        -------
        Optional[str]
            Returns a string describing the result of the command's execution with details if it failed.
        """
        return self._message        
    

# A composite command contains a command list but it behaves like a regular command to any other object using it.
# This means a composite command can contain a composite command and it can have many levels arbitrarily deep
# This also means it can potentially be recursive causing an infinite loop
# This also means that the contents of a composite command (and all potential levels of composite commands it has) should ideally be checked
# To make sure it does not cause problems for the overall recipe
# For now, CompositeCommands do not implement checks against adding other composite commands to itself as it can be useful to have at least 1-2 levels 
class CompositeCommand(Command):
    """A composite command which contains multiple commands but can act like a single command that executes all contained commands sequentially."""

    # receiver_cls = None
    
    def __init__(self, delay: float = 0.0):
        # super().__init__(**kwargs)
        # self._name = "CompositeCommand"
        # self._receiver_name = "N/A temp"
        self._command_list = []
        # self._receiver = None
        self._params = {}
        # self._params['receiver_name'] = 'None'
        self._params['delay'] = delay
        # self._was_successful = None
        # self._result_message = None  
        self._result = CommandResult()

    @property
    def name(self) -> str:
        """The composite command's name getter.

        Returns
        -------
        str
            Indicates it is a composite command followed by the names and params of each command it contains.
        """
        self._name = type(self).__name__ + " (CompositeCommand):"
        for command in self._command_list:
            self._name += " \n\t" + command.name + ";"
        return self._name

    @property
    def description(self) -> str:
        """The composite command's description getter.

        Returns
        -------
        str
            Returns the composite command's docstring
        """
        return self.__doc__

    def add_command(self, command: Command, index: Optional[int] = None):
        """Add a command to the command list.

        Parameters
        ----------
        command : Command
            Any command that subclasses from Command
        index : int, optional
            The index to insert the command, if None then the command is appended, by default None
        """
        if index is None:
            self._command_list.append(command)
        else: 
            self._command_list.insert(index, command)

    def remove_command(self, index: Optional[int] = None):
        """Remove a command from the command list.

        Parameters
        ----------
        index : int, optional
            The index of the command to remove, if None then removes the last command, by default None
        """
        if index is None:
            index = -1
        del self._command_list[index]

    def execute(self) -> None:
        """Executes each command in the command list sequentially and returns early if a command's execution was not successful."""
        # LOGGING? 
        for command in self._command_list:
            
            delay = command._params['delay']
            if isinstance(delay, float) or isinstance(delay, int):
                if delay > 0.0:
                    time.sleep(delay)
            elif delay == "PAUSE" or delay == "P":
                userinput = input()

            command.execute()
            # self._result.was_successful = command.was_successful
            # self._result._message = command.result_message
            self._result = command.result
            if not command.was_successful:
                return 

        # store the success of the last command 
        # or write a new message specific to the composite command
        # self._result_message = "Successfully executed composite command " + type(self).__name__


# for the receiver and concretecommand classes, I figured there were two ways of writing the code. Keep the receiver methods low level and employ logic 
# in the commands, or put all low level methods, logic, and subsequent high level methods in the receiver and have the command simply call the 
# high level methods with minimal logic if at all. I decided the latter mainly because some devices/receivers will have many different commands
# and I want to keep all the command classes streamlined and minimal. I also felt that having the logic and higher level methods separate from the lower level methods
# would make it more difficult to write and maintain due to switching back and forth between the modules.
# If there is a need to separate low level and high level methods, then in my opinion it should take place between the receiver class and another class it extends/inherits 
# from or the actual device firmware.

# nearly all attributes in Command and its child classes are protected because the parameters are not meant to be changed after construction.
# They can be, but for now I would just suggest deleting existing commands that needs to be changed and creating a new one in its place
# The initial reasoning for this is that the name and description may not update when params are changed because they are set during construction
# This can be fixed by implementing the arg dictionary mentioned below and using the name and description getters to update the value by iterating the dict
# This parallels the way CompositeCommands have names that update to reflect its command list when the name getter is used

# consider a command rank system to be able to check that some commands must come before/after others, e.g. initialize?


import threading
import random
from devices.device import MiscDeviceClass

class DummyMotorSource(MiscDeviceClass):

    def __init__(self, speed: float = 20.0):
        self._speed = speed
        self.min_speed = 5.0
        self.max_speed = 50.0
        self.min_position = 0.0
        self.max_position = 100.0
        self._position = random.uniform(self.min_position, self.max_position)
        self._hardware_interval = 0.05

    @property
    def position(self) -> float:
        return self._position

    @property
    def speed(self) -> float:
        return self._speed

    @speed.setter
    def speed(self, speed: float):
        if speed >= self.min_speed and speed <= self.max_speed:
            self._speed = speed
    
    def home_motor(self):
        self.move_absolute(position=0.0)

    # Move functions do not wait to reach target, it simply initiates the move
    def move_absolute(self, position: float):
        distance = position - self._position
        if distance > 0.0:
            self._hardware_move(position, 1)
        elif distance < 0.0:
            self._hardware_move(position, -1)

    def move_relative(self, distance: float):
        position = self._position + distance
        if distance > 0.0:
            self._hardware_move(position, 1)
        elif distance < 0.0:
            self._hardware_move(position, -1)       

    # Emulates change in motor position
    def _hardware_move(self, target: float, direction: int):
        # direction is 1 or -1
        timer = threading.Timer(self._hardware_interval, self._hardware_move, [target, direction])
        timer.start()
        distance_moved = self._speed * self._hardware_interval * direction

        if abs(target - self._position) <= abs(distance_moved):
            self._position = target
            timer.cancel()
        else:
            self._position += distance_moved
        


from .command import Command, CommandResult
from devices.multi_stepper import MultiStepper


class MultiStepperParentCommand(Command):
    """Parent class for all MultiStepper commands."""
    receiver_cls = MultiStepper

    def __init__(self, receiver: MultiStepper, **kwargs):
        super().__init__(receiver, **kwargs)

class MultiStepperConnect(MultiStepperParentCommand):
    """Open the one serial port for all stepper's controlled by a single arduino."""
    
    def __init__(self, receiver: MultiStepper, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.start_serial())
        
class MultiStepperInitialize(MultiStepperParentCommand):
    """Initialize all passed steppers by homing them."""
    
    def __init__(self, receiver: MultiStepper, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class MultiStepperDeinitialize(MultiStepperParentCommand):
    """Deinitialize all passed steppers by zeroing them."""

    def __init__(self, receiver: MultiStepper, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))

class MultiStepperMoveAbsolute(MultiStepperParentCommand):
    """Move stepper to absolute position."""

    def __init__(self, receiver: MultiStepper, stepper_number: int, position: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['stepper_number'] = stepper_number
        self._params['position'] = position

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_absolute(self._params['stepper_number'], self._params['position']))

class MultiStepperMoveRelative(MultiStepperParentCommand):
    """Move stepper by relative distance."""

    def __init__(self, receiver: MultiStepper, stepper_number: int, distance: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['stepper_number'] = stepper_number
        self._params['distance'] = distance

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.move_relative(self._params['stepper_number'], self._params['distance']))


# import pymongo
import yaml
from pymongo import MongoClient

# # Connection parameters
# host = 'localhost'  # Replace with your MongoDB server's hostname or IP address
# port = 27017  # Replace with the MongoDB server's port number

# # Create a MongoClient object
# client = MongoClient(host, port)


class MongoDBHelper:
    def __init__(self, host, port, database):
        """
        Initializes a MongoDBHelper object.

        Args:
            host (str): The MongoDB server's hostname or IP address.
            port (int): The MongoDB server's port number.
            database (str): The name of the MongoDB database to connect to.
        """
        self.client = MongoClient(host, port)
        self.db = self.client[database]

    def __init__(self, uri, database):
        """
        Initializes a MongoDBHelper object.

        Args:
            uri (str): The MongoDB connection URI string.
            database (str): The name of the MongoDB database to connect to.
        """
        self.client = MongoClient(uri)
        self.db = self.client[database]

    def insert_document(self, collection, document):
        """
        Inserts a document into the specified collection.

        Args:
            collection (str): The name of the collection to insert the document into.
            document (dict): The document to be inserted.

        Returns:
            pymongo.results.InsertOneResult: The result of the insertion operation.
        """
        return self.db[collection].insert_one(document)

    def insert_yaml_file(self, collection, file_path):
        """
        Inserts a YAML file into the MongoDB collection.

        Args:
            collection (str): The name of the collection to insert the document into.
            file_path (str): The path to the YAML file.

        Returns:
            str: The inserted document ID.
        """
        with open(file_path, 'r') as file:
            # yaml_data = yaml.load(file, Loader=yaml.Loader)
            yaml_data = file.read()
            doc = {'yaml_data': yaml_data, 'file_name': file_path}

        return str(self.db[collection].insert_one(doc).inserted_id)
    
    def update_yaml_file(self, collection, file_id, updated_data):
        """
        Updates a YAML file in the MongoDB collection.

        Args:
            collection (str): The name of the collection to insert the document into.
            file_id (str): The ID of the document representing the YAML file.
            updated_data (dict): The updated YAML data.

        Returns:
            bool: True if the update is successful, False otherwise.
        """
        result = self.db[collection].update_one({"_id": file_id}, {"$set": updated_data})
        return result.modified_count > 0
    
    def get_yaml_file(self, collection, file_id):
        """
        Retrieves a YAML file from the MongoDB collection.

        Args:
            collection (str): The name of the collection to insert the document into.
            file_id (str): The ID of the document representing the YAML file.

        Returns:
            dict: The YAML data.
        """
        return self.db[collection].find_one({"_id": file_id})

    def find_documents(self, collection, query):
        """
        Finds documents in the specified collection that match the query.

        Args:
            collection (str): The name of the collection to search.
            query (dict): The query to filter the documents.

        Returns:
            pymongo.cursor.Cursor: The cursor object containing the matching documents.
        """
        return self.db[collection].find(query)

    def update_documents(self, collection, filter_query, update_query):
        """
        Updates multiple documents in the specified collection.

        Args:
            collection (str): The name of the collection to update.
            filter_query (dict): The filter query to select the documents to update.
            update_query (dict): The update query specifying the changes to apply.

        Returns:
            pymongo.results.UpdateResult: The result of the update operation.
        """
        return self.db[collection].update_many(filter_query, update_query)

    def delete_documents(self, collection, filter_query):
        """
        Deletes multiple documents from the specified collection.

        Args:
            collection (str): The name of the collection to delete documents from.
            filter_query (dict): The filter query to select the documents to delete.

        Returns:
            pymongo.results.DeleteResult: The result of the delete operation.
        """
        return self.db[collection].delete_many(filter_query)

    def close_connection(self):
        """
        Closes the connection to the MongoDB server.
        """
        self.client.close()



# Examples

# # Create an instance of MongoDBHelper
# mongo = MongoDBHelper('localhost', 27017, 'your_database_name')

# # Example usage of insert_document()
# document = {'name': 'John', 'age': 30}
# result = mongo.insert_document('your_collection_name', document)
# # Document inserted. result.inserted_id contains the ID of the inserted document.

# # Example usage of find_documents()
# query = {'name': 'John'}
# documents = mongo.find_documents('your_collection_name', query)
# # documents is a cursor object containing the documents that match the query.
# # Iterate over the cursor to access each document.

# # Example usage of update_documents()
# filter_query = {'name': 'John'}
# update_query = {'$set': {'age': 31}}
# result = mongo.update_documents('your_collection_name', filter_query, update_query)
# # Multiple documents updated. result.modified_count contains the number of modified documents.

# # Example usage of delete_documents()
# filter_query = {'name': 'John'}
# result = mongo.delete_documents('your_collection_name', filter_query)
# # Multiple documents deleted. result.deleted_count contains the number of deleted documents.

# # Close the connection to the MongoDB server
# mongo.close_connection()







import setuptools


with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="aamp_app",  # This is the name of the package
    version="0.0.1.7",  # The initial release version
    author="Piyush Pahuja",  # Full name of the author
    description="AAMP App",
    long_description=long_description,  # Long description read from the the readme file
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),  # List of all python modules to be installed
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],  # Information to filter the project on PyPi website
    python_requires=">=3.6",  # Minimum version requirement of the package
    py_modules=[
        "aamp_app",
        "command_sequence",
        "command_invoker",
    ],  # Name of the python package
    package_dir={"aamp_app": "src"},  # Directory of the source code of the package
    install_requires=[
        "certifi",
        "colorama",
        "dash",
        "dash-ace",
        "dash-bootstrap-components",
        "dash-table",
        "mfc",
        "pandas",
        "pdoc",
        "Pillow",
        "pymongo",
        "pyserial",
        "PyVISA",
        "pyyaml",
        "questionary",
        "scipy",
        "slackclient",
        "ximea-py",
        "pyfirmata",
    ],  # Install other dependencies if any
)



from typing import Union
import time

try:
    import slack
    from slack.errors import SlackApiError
    import os
except ImportError:
    _has_slack = False
else:
    _has_slack = True

    
from .command import Command, CommandResult

class UtilityParentCommand(Command):
    """Parent class for utility commands that performs some function for its execute method but does not actually have a receiver."""

    # receiver_cls = None

    def __init__(self, delay: float = 0.0):
        # self._receiver = None
        self._params = {}
        # self._params['receiver_name'] = 'None'
        self._params['delay'] = delay
        # self._was_successful = None
        # self._result_message = None  
        self._result = CommandResult()

class LoopStartCommand(UtilityParentCommand):
    """Marks the start of a loop."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def execute(self) -> None:
        self._result = CommandResult(True, "Currently at loop start.")

class LoopEndCommand(UtilityParentCommand):
    """Marks the end of a loop."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def execute(self) -> None:
        self._result = CommandResult(True, "Currently at loop end.")

class DelayPauseCommand(UtilityParentCommand):
    """Has a delay or pause but does nothing."""

    def __init__(self, delay: Union[str, float]):
        arg_dict={}
        arg_dict['delay'] = delay
        super().__init__(**arg_dict)

    def execute(self) -> None:
        self._result = CommandResult(True, "Does nothing.")

class NotifySlackCommand(UtilityParentCommand):
    """Send a message to a designated slack channel"""

    def __init__(self, message: str, **kwargs):
        super().__init__(**kwargs)
        self._params['message'] = message.replace(" ", "_") # just for parsing name with spaces
        self._slack_message = message
        self._slack_token = os.environ.get('SLACK_BOT_TOKEN')
        self._slack_client = slack.WebClient(token=self._slack_token)

    def execute(self) -> None:
        try:
            response = self._slack_client.chat_postMessage(
                channel="printer-bot-test",
                text=("from NotifySlackCommand: " + self._slack_message)
                )
            self._result = CommandResult(True, "Successfully sent message.")  
        except SlackApiError as inst:
            self._result = CommandResult(False, "Could not send message: " + inst.response['error'])  

class LogUserMessageCommand(UtilityParentCommand):
    """Store a message in the command's result.message so it can be logged during invocation."""

    def __init__(self, message: str, **kwargs):
        super().__init__(**kwargs)
        self._params['message'] = message.replace(" ", "_") # just for parsing name with spaces
        # self._result_message = message
    
    def execute(self):
        self._result = CommandResult(True, self._params['message'])





from .command import Command, CommandResult, CompositeCommand
from devices.dummy_heater import DummyHeater

class DummyHeaterParentCommand(Command):
    """Parent class for all DummyHeater commands."""
    receiver_cls = DummyHeater

    def __init__(self, receiver: DummyHeater, **kwargs):
        super().__init__(receiver, **kwargs)

class DummyHeaterInitialize(DummyHeaterParentCommand):
    """Initialize the heater by setting to room temp (25 C)."""
    
    def __init__(self, receiver: DummyHeater, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class DummyHeaterDeinitialize(DummyHeaterParentCommand):
    """Deinitialize the heater by setting to room temp (25 C)."""
    
    def __init__(self, receiver: DummyHeater, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))

class DummyHeaterSetHeatRate(DummyHeaterParentCommand):
    """Set the heating rate of the heater."""

    def __init__(self, receiver: DummyHeater, heat_rate: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['heat_rate'] = heat_rate

    def execute(self) -> None:
        self._receiver.heat_rate = self._params['heat_rate']

        # logic implemented here because heat_rate setter does not return Tuple[bool, str]
        if self._receiver.heat_rate == self._params['heat_rate']:
            self._result = CommandResult(True, "Successfully set DummyHeater heat rate to " + str(self._receiver.heat_rate))
        else:
            self._result = CommandResult(False, "Failed to set DummyHeater heat rate. Heat rate is currently " + str(self._receiver.heat_rate))

class DummyHeaterSetTemp(DummyHeaterParentCommand):
    """Set the temperature of the heater and wait for it to stabilize."""

    def __init__(self, receiver: DummyHeater, temperature: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['temperature'] = temperature

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_temp(self._params['temperature']))

# Composite command example
class DummyHeaterRampHoldRamp(CompositeCommand):
    """Ramps up to temperature1 @ rate1, holds for specified time, then ramps to temperature2 @ rate2"""

    def __init__(
            self, 
            receiver: 
            DummyHeater, 
            temp1: float, 
            rate1: float, 
            hold_time: float, 
            temp2: float, 
            rate2: float,
            **kwargs):
        super().__init__(**kwargs)
        self.add_command(DummyHeaterSetHeatRate(receiver, rate1))
        self.add_command(DummyHeaterSetTemp(receiver, temp1))
        self.add_command(DummyHeaterSetHeatRate(receiver, rate2, delay=hold_time))
        self.add_command(DummyHeaterSetTemp(receiver, temp2))



from .command import Command, CommandResult
from devices.heating_stage import HeatingStage

#parent class for all HeatingStage commands
class HeatingStageParentCommand(Command):
    """Parent class for all HeatingStage commands."""
    receiver_cls = HeatingStage

    def __init__(self, receiver: HeatingStage, **kwargs):
        super().__init__(receiver, **kwargs)

class HeatingStageConnect(HeatingStageParentCommand):
    """Open serial port of heating stage's arduino controller."""
    
    def __init__(self, receiver: HeatingStage, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        # the serial port parameters should already be set in the HeatingStage instance
        self._result = CommandResult(*self._receiver.start_serial())

class HeatingStageInitialize(HeatingStageParentCommand):
    """Initialize heating stage by setting to room temperature and turning PID ON."""
    
    def __init__(self, receiver: HeatingStage, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class HeatingStageDeinitialize(HeatingStageParentCommand):
    """Deinitialize heating stage by setting to room temperature and turning PID OFF."""
    
    def __init__(self, receiver: HeatingStage, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))


class HeatingStageSetTemp(HeatingStageParentCommand):
    """Set heating stage temperature. Execution completes when stabilized."""
    
    def __init__(self, receiver: HeatingStage, temperature: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['temperature'] = temperature

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_temp(self._params['temperature']))

class HeatingStageSetSetPoint(HeatingStageParentCommand):
    """Set heating stage setpoint temperature. Execution completes immediately after setting setpoint, even if temperature is not stabilized."""
    
    def __init__(self, receiver: HeatingStage, temperature: float, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['temperature'] = temperature

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.set_settemp(self._params['temperature']))



from abc import ABC, abstractmethod
from typing import Optional, Tuple
import time
import functools
try:
    import serial
except ImportError:
    pass
import json


# Decorator to check if is initialized, optional custom message on fail
# Originally in Device class as a method, but moved to module scope
def check_initialized(func=None, *, message=None):
    if func is None:
        return functools.partial(check_initialized, message=message)

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self._is_initialized:
            if message is None:
                return (False, type(self).__name__ + " object is not initialized.")
            else:
                return (False, message)
        return func(self, *args, **kwargs)
    return wrapper

# consider optional flag to set is_initialized to False on fail
# this is so if this decorator is applied to the initialize function
# the function will have a chance to set is_initialized to False before the wrapper returns
# Note that in some cases the initialize function starts by calling another function that already checks serial
def check_serial(func=None, *, message=None):
    if func is None:
        return functools.partial(check_initialized, message=message)

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not self.ser.is_open:
            if message is None:
                return (False, "Serial port " + self._port + " is not open.")
            else:
                return (False, message)
        return func(self, *args, **kwargs)
    return wrapper


# *args, **kwargs https://stackoverflow.com/questions/6034662/python-method-overriding-does-signature-matter
# Consider using *args and **kwargs, although you lose the ability of the IDE to hint at the args needed in a signature

class Device(ABC):
    """The Device abstract base class that contains attributes that all devices/receivers should have."""

    def __init__(self, name: str):
        self._name = name
        self._is_initialized = False

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def is_initialized(self) -> bool:
        """Whether or not the device has been initialized.

        Returns
        -------
        bool
            Returns True if the device is initialized, otherwise returns False.
        """
        return self._is_initialized

    @abstractmethod
    def initialize(self) -> Tuple[bool, str]:
        """The initialize abstract method that all devices should implement.

        Returns
        -------
        Tuple[bool, str]
            Returns a bool describing whether the initialization was successful and a string describing the result.
        """
        pass

    @abstractmethod
    def deinitialize(self) -> Tuple[bool, str]:
        """The deinitialize abstract method that all devices should implement.

        Returns
        -------
        Tuple[bool, str]
            Returns a bool describing whether the deinitialization was successful and a string describing the result.
        """
        pass

    @abstractmethod # uncomment and implement method in all devices
    def get_init_args(self) -> dict:
        """The get_args abstract method that all devices should implement. Method should return a dict with only the arguments needed to initialize the device.

        Returns
        -------
        dict
            Returns a dict containing the arguments needed to initialize the device.
        """
        pass

    @abstractmethod # uncomment and implement method in all devices
    def update_init_args(self, args_dict: dict):
        """The update_init_args abstract method that all devices should implement. Method should update the arguments needed to initialize the device.
        
        Parameters
        ----------
        
        args_dict : dict
            A dict containing the arguments needed to initialize the device.
        """
        pass


class SerialDevice(Device):
    """A Device that uses serial communication."""

    def __init__(self, name: str, port: str, baudrate: int, timeout: Optional[float] = 1.0):
        super().__init__(name)
        self._port = port
        self._baudrate = baudrate
        self._timeout = timeout
        self.ser = serial.Serial()

    @property
    def port(self) -> str:
        return self._port
    
    @port.setter
    def port(self, port: str):
        self._port = port

    @property
    def baudrate(self) -> int:
        return self._baudrate
    
    @baudrate.setter
    def baudrate(self, baudrate: int):
        self._baudrate = baudrate

    @property
    def timeout(self) -> Optional[float]:
        return self._timeout
    
    @timeout.setter
    def timeout(self, timeout: Optional[float]):
        self._timeout = timeout

    def start_serial(self, delay: float = 5.0) -> Tuple[bool, str]:
        """Set the serial port parameters and try to open the serial port

        Parameters
        ----------
        delay : float, optional
            How long to wait after opening the serial port before returning, by default 5.0

        Returns
        -------
        Tuple[bool, str]
            Was the serial port opened?, Result message
        """
        # self.ser.port = self._port
        # self.ser.baudrate = self._baudrate
        # self.ser.timeout = self._timeout
        try:
            if not self.ser.is_open:
                self.ser.port = self._port
                self.ser.baudrate = self._baudrate
                self.ser.timeout = self._timeout
                self.ser.open()   
                # delay at least 5s for arduino
                time.sleep(delay)  
                return (True, "Serial port " + self._port + " successfully opened.") 
            # If setting self.ser... before try, then it will restart an arduino and maybe other devices
            # time.sleep(delay)  
            return (True, "Serial port " + self._port + " already opened.")
        except serial.serialutil.SerialException as inst:
            return (False, "Serial port " + self._port + " failed to open. " + str(inst))     
  

class ArduinoSerialDevice(SerialDevice):
    """An arduino serial device that implements the custom ACK/NACK/SUCC/FAIL communication protocol."""

    char_ACK = "ACK"
    char_SUCC = "SUCC"
    char_delimiter = ":"
    partial_timeout = 2.0 # time in seconds to wait if a partial message is received but not newline footer

    def __init__(self, name: str, port: str, baudrate: int, timeout: Optional[float] = 1.0):
        super().__init__(name, port, baudrate, timeout)

    def check_ack_succ(self, ack_timeout: float = 1.0, succ_timeout: float = 1.0) -> Tuple[bool, str]:
        """Checks whether the instruction was acknowledged and successful.

        Parameters
        ----------
        ack_timeout : float, optional
            How long to wait for ACK/NACK response, by default 1.0
        succ_timeout : float, optional
            How long to wait for SUCC/FAIL response, by default 1.0

        Returns
        -------
        Tuple[bool, str]
            Was the instruction acknowledged and succesful?, Result message
        """
        was_acknowledged, ack_comment = self.check_response(
            ArduinoSerialDevice.char_ACK, 
            ArduinoSerialDevice.char_delimiter, 
            ack_timeout)

        if not was_acknowledged:
            return (was_acknowledged, ack_comment)

        return self.check_response(
            ArduinoSerialDevice.char_SUCC, 
            ArduinoSerialDevice.char_delimiter, 
            succ_timeout)
   
    # in future may alter control_char to be of type Union[str, bytes] and be able to process a byte control character
    def check_response(self, control_char: str, delimiter: str, response_timeout: float = 1.0) -> Tuple[bool, str]:
        """Checks whether a response with a particular starting string was received.

        Parameters
        ----------
        control_char : str
            The string that is being checked for
        delimiter : str
            The delimiter between the control_char string and any additional information or data.
        response_timeout : float, optional
            How long to wait for the response, by default 1.0

        Returns
        -------
        Tuple[bool, str]
            Was the desired control_char string received?, Result message
        """
        was_successful, message = self.get_response(response_timeout)
        if not was_successful:
            return (False, message)
        else:
            response_result = message
        
        if delimiter in response_result:
            # the response comes with a message/comment
            response_char = response_result.split(delimiter)[0].strip()
            response_comment = response_result.split(delimiter)[1].strip()
        else:
            # the response does not come with a message
            response_char = response_result.strip()
            response_comment = ""

        if response_char != control_char:
            return (False, "Did not receive the control char or str " + control_char + ". Instead received: " + response_result + ".")

        return (True, "Successfully received the control char or str " + control_char + " with message: " + response_comment + ".")

    def get_response(self, response_timeout: float = 1.0) -> Tuple[bool, str]:
        retry_count = 0
        partial_retries = ArduinoSerialDevice.partial_timeout // self._timeout
        response_retries = response_timeout // self._timeout

        if response_retries < 1:
            response_retries = 1
        
        # using integer retries of period self.timeout, this is because changing ser.timeout directly causes problems
        response_result = b''
        while response_result.decode('ascii') == "" and retry_count < response_retries:
            response_result = self.ser.readline()
            retry_count += 1

        # https://stackoverflow.com/questions/61166544/readline-in-pyserial-sometimes-captures-incomplete-values-being-streamed-from
        # in case readline times out in the middle of a message before \n
        retry_count = 0
        while response_result.decode('ascii') != "" and "\\n" not in str(response_result) and retry_count < partial_retries:
            temp_result = self.ser.readline()
            retry_count += 1
            
            # "not not" is correct, means not empty, != ""
            if not not temp_result.decode('ascii'):
                response_result = (response_result.decode('ascii') + temp_result.decode('ascii')).encode('ascii')
        
        if retry_count == partial_retries and "\\n" not in str(response_result):
            return (False, "Timed out. Partial message received for potential control char or str " + control_char + ".")

        response_result = response_result.strip().decode('ascii')

        if response_result == "":
            return (False, "Timed out. Did not receive any response for control char or str " + control_char + ".")

        return (True, response_result)
    
    @staticmethod
    def parse_equal_sign(text: str) -> Tuple[bool, str]:
        """For strings that contain information, get the desired information that is located at the end of the string which follows an = sign.

        Parameters
        ----------
        text : str
            the string that contains an = sign followed by the desired information.

        Returns
        -------
        Tuple[bool, str]
            Whether the string contained an = sign, and the desired information as a string.
        """
        if "=" in text:
            last_token = text.split("=")[-1]
            return (True, last_token.strip())
        else:
            return (False, "")
        

class MiscDeviceClass():
    def exists():
        return True


from .command import Command, CommandResult
from devices.dummy_meter import DummyMeter

class DummyMeterParentCommand(Command):
    """Parent class for all DummyMeter commands."""
    receiver_cls = DummyMeter

    def __init__(self, receiver: DummyMeter, **kwargs):
        super().__init__(receiver, **kwargs)

class DummyMeterInitialize(DummyMeterParentCommand):
    """Initialize the meter."""
    
    def __init__(self, receiver: DummyMeter, **kwargs):
        super().__init__(receiver, **kwargs)

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.initialize())

class DummyMeterDeinitialize(DummyMeterParentCommand):
    """Deinitialize the meter."""
    
    def __init__(self, receiver: DummyMeter, reset_init_flag: bool = True, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['reset_init_flag'] = reset_init_flag

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.deinitialize(self._params['reset_init_flag']))

class DummyMeterMeasure(DummyMeterParentCommand):
    """Measure data and save to file."""

    def __init__(self, receiver: DummyMeter, filename: str = None, **kwargs):
        super().__init__(receiver, **kwargs)
        self._params['filename'] = filename

    def execute(self) -> None:
        self._result = CommandResult(*self._receiver.measure(self._params['filename']))